"""
### Align rasters ###

Functions to align a series of rasters to a master or a reference.
"""

# TODO: phase_cross_correlation

# Standard library
import sys; sys.path.append("../../")
from typing import List, Union, Optional
from uuid import uuid4

# External
from osgeo import gdal

# Internal
from buteo.utils import core_utils, gdal_enums, gdal_utils, bbox_utils
from buteo.raster import core_raster
from buteo.raster.reproject import reproject_raster, match_raster_projections


def align_rasters_to_reference(
    rasters: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    reference: Union[str, gdal.Dataset],
    *,
    out_path: Optional[Union[str, List[str]]] = None,
    resample_alg: str = "nearest",
    target_nodata: Optional[Union[int, float]] = None,
    overwrite=True,
    creation_options: Optional[List[str]] = None,
    prefix="",
    suffix="",
    ram="auto",
) -> List[str]:
    """
    Aligns a series of rasters to a reference.

    Args:
        rasters (list): A list of rasters to align.
        reference (str/gdal.Dataset): Path to the reference raster
            or vector.

    Keyword Args:
        out_path (str/list, default=None): Paths to the output. If not provided,
            the output will be in-memory rasters.
        resample_alg (str, default="nearest"): Resampling algorithm to use.
        target_nodata (int/float, default=None): Nodata value to use for the
            output rasters.
        overwrite (bool, default=True): Overwrite existing files.
        creation_options (list, default=None): List of creation options.
        prefix (str, default=""): Prefix to add to the output file name.
        suffix (str, default=""): Suffix to add to the output file name.
        ram (int/str, default="auto"): Amount of RAM to use in MB. If "auto",
            the amount of RAM will be determined automatically.
    """
    core_utils.type_check(rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "rasters")
    core_utils.type_check(reference, [str, gdal.Dataset], "reference")
    core_utils.type_check(out_path, [str, None, [str]], "out_path")
    core_utils.type_check(resample_alg, [str], "resample_alg")
    core_utils.type_check(target_nodata, [int, float, None], "target_nodata")
    core_utils.type_check(overwrite, [bool], "overwrite")
    core_utils.type_check(creation_options, [[str], None], "creation_options")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "suffix")
    core_utils.type_check(ram, [int, str], "ram")

    rasters_list = core_utils.ensure_list(rasters)
    assert gdal_utils.is_raster_or_vector(reference), "Reference must be a raster or vector."

    add_uuid = True if out_path is None else False

    # Verify that all of the rasters overlap the reference
    for raster in rasters_list:
        assert core_raster.rasters_intersect(raster, reference), (
            f"Raster: {gdal_utils.get_path_from_dataset(raster)} does not intersect reference."
        )

    path_list = gdal_utils.create_output_path_list(
        rasters_list,
        out_path,
        overwrite=overwrite,
        add_uuid=add_uuid,
        ext=".tif",
        prefix=prefix,
        suffix=suffix,
    )

    creation_options = gdal_utils.default_creation_options(creation_options)

    reference_metadata = core_raster.raster_to_metadata(reference)

    target_projection = reference_metadata["projection_osr"]
    reference_bbox = reference_metadata["bbox"]
    x_min, _, _, y_max = reference_bbox

    # Set the target values.
    x_res = reference_metadata["pixel_width"]
    y_res = reference_metadata["pixel_height"]
    x_pixels = reference_metadata["width"]
    y_pixels = reference_metadata["height"]

    # Reproject the rasters to the reference projection.
    for idx, raster in enumerate(rasters_list):
        raster_metadata = core_raster.raster_to_metadata(raster)
        raster_path = gdal_utils.get_path_from_dataset(raster)

        raster_reprojected = None
        raster_projection = raster_metadata["projection_osr"]
        if not gdal_utils.projections_match(raster_projection, target_projection):
            raster_reprojected = reproject_raster(
                raster_path,
                target_projection,
                out_path=None,
                resample_alg=resample_alg,
                creation_options=creation_options,
                prefix="tmp_reprojection_",
                add_uuid=True,
            )
            raster_ds = core_raster.open_raster(raster_reprojected)
        else:
            raster_ds = core_raster.open_raster(raster_path)

        destination_ds = core_raster.create_empty_raster(
            out_path=path_list[idx],
            width=x_pixels,
            height=y_pixels,
            pixel_size=(x_res, y_res),
            x_min=x_min,
            y_max=y_max,
            dtype=raster_metadata["dtype"],
            bands=raster_metadata["band_count"],
            nodata_value=target_nodata,
            projection=target_projection,
            creation_options=creation_options,
            overwrite=overwrite,
        )

        destination_ds = core_raster.open_raster(destination_ds)

        warp_options = gdal.WarpOptions(
            resampleAlg=gdal_enums.translate_resample_method(resample_alg),
            multithread=True,
            outputBounds=bbox_utils.convert_ogr_bbox_to_gdal_bbox(reference_bbox),
            warpMemoryLimit=gdal_utils.get_gdalwarp_ram_limit(ram),
        )

        warped = gdal.Warp(destination_ds, raster_ds, options=warp_options)

        if raster_reprojected is not None:
            gdal_utils.delete_if_in_memory(raster_reprojected)

        destination_ds.FlushCache()
        destination_ds = None

        if warped is None:
            raise ValueError("Error while warping rasters.")

    return path_list


def find_best_reference(
    rasters: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    method: str,
) -> str:
    """ Find the best reference raster for aligning a list of rasters.
    
    Args:
        rasters (str/list): List of rasters to align.
        method (str): Bounding box method to use for finding the best reference.
            Options include: "reference", "intersection", and "union".
    Returns:
        str: Path to the best reference raster.
    """
    core_utils.type_check(rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "rasters")
    core_utils.type_check(method, [str], "method")

    rasters_list = core_utils.ensure_list(rasters)
    assert method in ["reference", "intersection", "union"], (
        f"Invalid bounding_box: {method}. Options include: reference, intersection, and union."
    )

    assert gdal_utils.is_raster_list(rasters_list), "All rasters must be raster datasets."

    if len(rasters_list) == 1:
        return rasters_list[0]

    # Count intersections
    most_intersections = 0
    intersections_arr = []
    for idx_i, raster_i in enumerate(rasters_list):
        intersections = 0
        for idx_j, raster_j in enumerate(rasters_list):
            if idx_i == idx_j:
                continue

            # Uses the latlng bounding box to determine if the rasters intersect.
            intersection = core_raster.rasters_intersect(raster_i, raster_j)

            if intersection:
                intersections += 1

        if intersections > most_intersections:
            most_intersections = intersections

        intersections_arr.append(intersections)

    # Find the raster with the most intersections.
    largest_area = 0.0
    largest_area_idx = 0
    for idx, intersection in enumerate(intersections_arr):
        if intersection == most_intersections:
            raster = rasters_list[idx]
            raster_metadata = core_raster.raster_to_metadata(raster)
            raster_area = raster_metadata["area_latlng"]

            if raster_area > largest_area:
                largest_area = raster_area
                largest_area_idx = idx

    best_reference = gdal_utils.get_path_from_dataset(rasters_list[largest_area_idx])

    input_path_list = []
    for raster in rasters_list:
        input_path_list.append(gdal_utils.get_path_from_dataset(raster))

    all_other_rasters = []
    for raster_path in input_path_list:
        if raster_path == best_reference:
            continue

        all_other_rasters.append(raster_path)

    for raster in all_other_rasters:
        assert core_raster.rasters_intersect(best_reference, raster), (
            f"Rasters {best_reference} and {raster} do not intersect."
        )

    if method == "reference":
        return best_reference

    matched_rasters = match_raster_projections(
        all_other_rasters,
        best_reference,
        copy_if_already_correct=False,
    )

    best_reference_meta = core_raster.raster_to_metadata(best_reference)

    best_geom = best_reference_meta["geom"]
    for raster_path in matched_rasters:
        raster_meta = core_raster.raster_to_metadata(raster_path)
        raster_geom = raster_meta["geom"]

        if not raster_geom.Intersects(best_geom):
            raise ValueError(
                f"Raster {raster_path} did not intersect. Consider not using bbox='intersection'"
            )

        if method == "union":
            best_geom = raster_geom.Union(best_geom)
        else:
            best_geom = raster_geom.Intersection(best_geom)

    bbox_best_geom = best_geom.GetEnvelope()

    for raster_path in matched_rasters:
        if raster_path not in input_path_list:
            gdal_utils.delete_if_in_memory(raster_path)

    outbox_bbox = bbox_utils.align_bboxes_to_pixel_size(
        best_reference_meta["bbox"],
        bbox_best_geom,
        best_reference_meta["pixel_width"],
        best_reference_meta["pixel_height"],
    )

    creation_options = gdal_utils.default_creation_options()

    width = int(round((outbox_bbox[1] - outbox_bbox[0]) / best_reference_meta["pixel_width"]))
    height = int(round((outbox_bbox[3] - outbox_bbox[2]) / best_reference_meta["pixel_height"]))

    out_path = core_raster.create_empty_raster(
        out_path=f"/vsimem/intersection_union_{uuid4().int}.tif",
        width=width,
        height=height,
        pixel_width=best_reference_meta["pixel_width"],
        pixel_height=best_reference_meta["pixel_height"],
        bands=1,
        dtype=best_reference_meta["dtype"],
        x_min=outbox_bbox[0],
        y_max=outbox_bbox[3],
        projection=best_reference_meta["projection"],
        creation_options=creation_options,
    )

    return out_path


def align_rasters(
    rasters: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    *,
    out_path: Optional[Union[str, List[str]]] = None,
    reference: Optional[Union[str, gdal.Dataset]] = None,
    method: str = "reference",
    target_nodata: Optional[Union[int, float]] = None,
    resample_alg: str = "nearest",
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
    prefix="",
    suffix="",
    ram: Union[int, str] = "auto",
) -> List[str]:
    """ Aligns rasters either to a reference raster or to each other using the
    one of three methods: reference, intersection, or union.
    
    Args:
        rasters (Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]]): The rasters to align.

    Keyword Args:
        out_path (Optional[Union[str, List[str]]], optional): The output path(s). Defaults to None.
        reference (Optional[Union[str, gdal.Dataset]], optional): The reference raster to align to. Defaults to None.
        method (str, optional): The method to use. Defaults to "reference" [reference, intersection, union].
        resample_alg (str, optional): The resampling algorithm to use. Defaults to "nearest".
        overwrite (bool, optional): Whether to overwrite existing files. Defaults to True.
        creation_options (Optional[List[str]], optional): The creation options to use. Defaults to None.
        prefix (str, optional): The prefix to add to the output file name. Defaults to "".
        suffix (str, optional): The suffix to add to the output file name. Defaults to "".
        ram (int/str, default="auto"): Amount of RAM to use in MB. If "auto",
            the amount of RAM will be determined automatically.
    
    Returns:
        List[str]: The aligned rasters.
    """
    core_utils.type_check(rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "rasters")
    core_utils.type_check(out_path, [str, None, [str]], "out_path")
    core_utils.type_check(reference, [str, [str], None], "reference")
    core_utils.type_check(method, [str], "method")
    core_utils.type_check(target_nodata, [int, float, None], "target_nodata")
    core_utils.type_check(resample_alg, [str], "resample_alg")
    core_utils.type_check(overwrite, [bool], "overwrite")
    core_utils.type_check(creation_options, [[str], None], "creation_options")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "suffix")
    core_utils.type_check(ram, [int, str], "ram")

    raster_list = core_utils.ensure_list(rasters)

    assert gdal_utils.is_raster_list(raster_list), "rasters must be a single raster or a list of rasters."
    assert method in ["reference", "intersection", "union"], "method must be one of reference, intersection, or union."

    add_uuid = True if out_path is None else False

    path_list = gdal_utils.create_output_path_list(
        raster_list,
        out_path,
        overwrite=overwrite,
        add_uuid=add_uuid,
        ext=".tif",
        prefix=prefix,
        suffix=suffix,
    )

    if reference is None:
        reference = find_best_reference(raster_list, method)

    aligned = align_rasters_to_reference(
        raster_list,
        reference,
        out_path=path_list,
        resample_alg=resample_alg,
        target_nodata=target_nodata,
        overwrite=overwrite,
        creation_options=creation_options,
        suffix=suffix,
        prefix=prefix,
        ram=ram,
    )

    return aligned



"""
### Perform convolutions on arrays.  ###
"""

# External
import numpy as np
from numpy.lib.stride_tricks import as_strided
from numba import jit, prange


def weight_distance(
    arr,
    method=None,
    decay=0.2,
    sigma=1.0,
    center=0.0,
    spherical=False,
    radius=3.0,
):
    """
    Weights the kernel by distance using various methods.

    Args:
        arr (numpy.ndarray): The input array.
        method (str, default=None): The weighting method to use.
            "none": No weighting (default).
            "linear": Linear decay.
            "sqrt": Square root decay.
            "power": Power decay.
            "log": Logarithmic decay.
            "gaussian": Gaussian decay.
        decay (float, default=0.2): The decay rate for the `linear`, `sqrt`, and `power` methods.
        sigma (float, default=1.0): The standard deviation for the Gaussian method.
        center (float, default=0.0): The center of the array.
        spherical (bool, default=False): If True, adjust weights based on the radius.
        radius (float, default=3.0): The radius for spherical adjustments.

    Returns:
        float: The computed weight.
    """

    if center == 0.0:
        normed = np.linalg.norm(arr)
    else:
        normed = np.linalg.norm(arr - np.array([0, 0, center]))

    if normed == 0.0:
        weight = 1.0

    if method is None or method == "none" or method == "":
        weight = 1.0
    elif method == "linear":
        weight = np.power((1 - decay), normed)
    elif method == "sqrt":
        weight = np.power(np.sqrt((1 - decay)), normed)
    elif method == "power":
        weight = np.power(np.power((1 - decay), 2), normed)
    elif method == "log":
        weight = np.log(normed + 2)
    elif method == "gaussian":
        weight = np.exp(-(np.power(normed, 2)) / (2 * np.power(sigma, 2)))
    else:
        raise ValueError("Unable to parse parameters distance_calc.")

    if spherical:
        sqrt_2 = np.sqrt(2)
        half_sqrt_2 = np.divide(sqrt_2, 2)

        if normed > (radius - half_sqrt_2) and normed < (radius + half_sqrt_2):
            adjustment = sqrt_2 / normed
        elif normed > (radius - half_sqrt_2):
            adjustment = 0.0
        elif normed < (radius + half_sqrt_2):
            adjustment = 1.0
        else:
            adjustment = 1.0

        return weight * adjustment

    return weight


def rotate_kernel(bottom_right):
    """
    Creates a whole kernel from a quadrant.

    Args:
        bottom_right (numpy.ndarray): The bottom-right quadrant of the kernel.

    Returns:
        numpy.ndarray: The complete kernel generated from the given quadrant.
    """

    size = ((bottom_right.shape[0] - 1) * 2) + 1
    depth = bottom_right.shape[2]
    kernel = np.zeros((size, size, depth), dtype="float32")

    top_right = np.flipud(bottom_right)
    lower_left = np.fliplr(bottom_right)
    top_left = np.flipud(lower_left)

    kernel = np.block([[top_left, top_right], [lower_left, bottom_right]])

    return kernel


def get_kernel(
    size,
    depth=1,
    hole=False,
    inverted=False,
    normalise=True,
    multi_dimensional=False,
    multi_dimensional_center=0,
    spherical=False,
    distance_weight=None,
    distance_decay=0.2,
    distance_sigma=1,
):
    """
    Generates a square kernel for convolutions.

    Args:
        size (int): Size of the kernel (must be odd).
        depth (int, default=1): Depth of the kernel.
        hole (bool, default=False): Create a hole in the center of the kernel.
        inverted (bool, default=False): Invert the kernel values.
        normalise (bool, default=True): Normalize the kernel values.
        multi_dimensional (bool, default=False): Consider the kernel multi-dimensional.
        multi_dimensional_center (int, default=0): Center of the
            multi-dimensional kernel.
        spherical (bool, default=False): Consider the kernel spherical.
        distance_weight (str or None, default=None): Distance weighting method.
        distance_decay (float, default=0.2): Distance decay factor.
        distance_sigma (float, default=1): Distance sigma for Gaussian distance weighting.

    Returns:
        tuple: A tuple containing the kernel, weights, and offsets.
    """

    assert size >= 3, "Kernel must have atleast size 3."
    assert size % 2 != 0, "Kernel must be an uneven size."
    assert isinstance(size, int), "Kernel must be an integer."
    assert depth >= 1, "Depth must be a positive integer"
    assert isinstance(depth, int), "Depth must be an integer."

    if distance_weight is False:
        distance_weight = None

    quadrant = np.zeros((1 + size // 2, 1 + size // 2, depth), dtype="float32")

    for idx_x in range(0, quadrant.shape[0]):
        for idx_y in range(0, quadrant.shape[1]):
            for idx_z in range(0, quadrant.shape[2]):

                z_value = idx_z if multi_dimensional else 0

                weighted = weight_distance(
                    np.array([idx_x, idx_y, z_value], dtype="float32"),
                    method=distance_weight,
                    decay=distance_decay,
                    sigma=distance_sigma,
                    center=multi_dimensional_center,
                    spherical=spherical,
                    radius=size // 2,
                )

                quadrant[idx_x, idx_y, idx_z] = weighted
    if hole:
        for idx_z in range(0, quadrant.shape[2]):
            quadrant[0, 0, idx_z] = 0

    kernel = rotate_kernel(quadrant)

    if distance_weight == "log":
        kernel = kernel.max() - kernel

    if inverted:
        kernel = 1 - kernel

    if normalise:
        if multi_dimensional:
            summed = kernel.sum()
            if summed != 0.0:
                kernel = kernel / summed
        else:
            summed = kernel.sum(axis=(0, 1))

            for dim in range(0, depth):
                kernel[:, :, dim] = kernel[:, :, dim] / summed[dim]

    weights = []
    offsets = []

    for idx_x in range(0, kernel.shape[0]):
        for idx_y in range(0, kernel.shape[1]):
            for idx_z in range(0, kernel.shape[2]):
                current_weight = kernel[idx_x][idx_y][idx_z]

                if current_weight <= 0.0:
                    continue

                offsets.append(
                    [
                        idx_x - (kernel.shape[0] // 2),
                        idx_y - (kernel.shape[1] // 2),
                        idx_z
                    ]
                )

                weights.append(current_weight)

    return kernel, np.array(weights, dtype="float32"), np.array(offsets, dtype=int)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_max(values, weights):
    """ Get the weighted maximum. """
    idx = np.argmax(np.multiply(values, weights))
    return values[idx]


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_min(values, weights):
    """ Get the weighted minimum. """
    max_val = values.max()
    adjusted_values = np.where(weights == 0.0, max_val, values)
    idx = np.argmin(np.divide(adjusted_values, weights + 1e-7))
    return values[idx]


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_sum(values, weights):
    """ Get the weighted sum. """
    return np.sum(np.multiply(values, weights))


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_mode(values, weights):
    """ Get the weighted sum. """
    values_ints = np.rint(values)
    unique = np.unique(values_ints)

    most_occured_value = 0
    most_occured_weight = -9999.9

    for unique_value in unique:
        cum_weight = 0
        for idx in range(values.shape[0]):
            if values_ints[idx] == unique_value:
                cum_weight += weights[idx]

        if cum_weight > most_occured_weight:
            most_occured_weight = cum_weight
            most_occured_value = unique_value

    return most_occured_value


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_contrast(values, weights):
    """ Get the local contrast. """
    max_val = values.max()
    adjusted_values = np.where(weights == 0.0, max_val, values)
    local_min = np.min(np.divide(adjusted_values, weights + 1e-7))
    local_max = np.max(np.multiply(values, weights))

    return np.abs(local_max - local_min)

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_quantile(values, weights, quant):
    """ Get the weighted median. """
    sort_mask = np.argsort(values)
    sorted_data = values[sort_mask]
    sorted_weights = weights[sort_mask]
    cumsum = np.cumsum(sorted_weights)
    intersect = (cumsum - 0.5 * sorted_weights) / cumsum[-1]
    return np.interp(quant, intersect, sorted_data)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_median_absolute_deviation(values, weights):
    """ Get the median absolute deviation """
    median = hood_quantile(values, weights, 0.5)
    absdeviation = np.abs(np.subtract(values, median))
    return hood_quantile(absdeviation, weights, 0.5)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_z_score(values, center_value, weights):
    """ Get the local z score ."""
    std = hood_standard_deviation(values, weights)
    mean = hood_sum(values, weights)

    return (center_value - mean) / std


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_z_score_mad(values, center_value, weights):
    """ Get the local z score calculated around the MAD. """
    mad_std = hood_median_absolute_deviation(values, weights) * 1.4826
    median = hood_quantile(values, weights, 0.5)

    return (center_value - median) / mad_std


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_standard_deviation(values, weights):
    "Get the weighted standard deviation. "
    summed = hood_sum(values, weights)
    variance = np.sum(np.multiply(np.power(np.subtract(values, summed), 2), weights))
    return np.sqrt(variance)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def k_to_size(size):
    """ Preprocess Sigma Lee limits. """
    return int(np.rint(-0.0000837834 * size ** 2 + 0.045469 * size + 0.805733))


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_sigma_lee(values, weights):
    """ Sigma lee SAR filter. """
    std = hood_standard_deviation(values, weights)
    selected_values = np.zeros_like(values)
    selected_weights = np.zeros_like(weights)

    sigma_mult = 1
    passed = 0
    attempts = 0
    ks = k_to_size(values.size)

    while passed < ks and attempts < 5:
        for idx, val in np.ndenumerate(values):
            if val >= std * sigma_mult or val <= -std * sigma_mult:
                selected_values[idx] = val
                selected_weights[idx] = weights[idx]
                passed += 1

        sigma_mult += 1
        attempts += 1

    if passed < ks:
        return hood_sum(values, weights)

    sum_of_weights = np.sum(selected_weights)

    if sum_of_weights == 0:
        return 0

    selected_weights = np.divide(selected_weights, sum_of_weights)

    return hood_sum(selected_values, selected_weights)


@jit(nopython=True, nogil=True)
def pad_array_view(arr, pad_size=1):
    """
    Create a padded view of an array using SAME padding.

    Args:
        arr (numpy.ndarray): The input array to be padded.
2
    Keyword Args:
        pad_size (int, default=1): The number of padding elements to add
            to each side of the array. Default is 1.

    Returns:
        numpy.ndarray: A padded view of the input array.
    """
    # Get original array shape and strides
    shape = arr.shape
    strides = arr.strides

    # Compute new shape and strides for the padded view
    new_shape = (shape[0] + 2 * pad_size, shape[1] + 2 * pad_size, shape[2])
    new_strides = (strides[0], strides[1], strides[2])

    # Create the padded view using as_strided
    padded_view = as_strided(arr, shape=new_shape, strides=new_strides)

    # Pad the edges using the SAME padding
    padded_view[:pad_size, :, :] = padded_view[pad_size:pad_size + 1, :, :]
    padded_view[-pad_size:, :, :] = padded_view[-pad_size - 1:-pad_size, :, :]
    padded_view[:, :pad_size, :] = padded_view[:, pad_size:pad_size + 1, :]
    padded_view[:, -pad_size:, :] = padded_view[:, -pad_size - 1:-pad_size, :]

    return padded_view


@jit(nopython=True, nogil=True)
def pad_array(arr, pad_size=1):
    """ Pad an array using SAME """
    # Core
    arr_padded = np.zeros((arr.shape[0] + int(pad_size * 2), arr.shape[1] + int(pad_size * 2), arr.shape[2]), dtype=arr.dtype)
    arr_padded[pad_size:-pad_size, pad_size:-pad_size, :] = arr

    # Corners
    arr_padded[0:pad_size, 0:pad_size, :] = arr[ 0,  0, :]
    arr_padded[-pad_size:, -pad_size:, :] = arr[-1, -1, :]
    arr_padded[0:pad_size, -pad_size:, :] = arr[ 0, -1, :]
    arr_padded[-pad_size:, 0:pad_size, :] = arr[-1,  0, :]

    # Sides
    for idx in range(0, pad_size):
        arr_padded[idx, pad_size:-pad_size, :] = arr[ 0,  :, :]
        arr_padded[-(idx + 1):, pad_size:-pad_size, :] = arr[-1,  :, :]
        arr_padded[pad_size:-pad_size, -(idx + 1), :] = arr[ :, -1, :]
        arr_padded[pad_size:-pad_size, idx, :] = arr[ :,  0, :]

    return arr_padded

METHOD_ENUMS = {
    "sum": 1,
    "mode": 2,
    "max": 3,
    "min": 4,
    "contrast": 5,
    "median": 6,
    "std": 7,
    "mad": 8,
    "z_score": 9,
    "z_score_mad": 10,
    "sigma_lee": 11,
}

@jit(nopython=True, parallel=True, nogil=False, fastmath=True, cache=True)
def convolve_array(
    arr,
    offsets,
    weights,
    method=1,
    nodata=False,
    nodata_value=-9999.9,
):
    """
    Convolve an image with a function.

    Args:
        arr (numpy.ndarray): The input array to convolve.
        offsets (list of tuples): The list of offsets for the neighborhood
            used in the convolution.
        weights (list): The list of weights used in the convolution.

    Keyword Args:
        method (int, default=1): The method to use for the convolution.
            1: hood_sum
            2: hood_mode
            3: hood_max
            4: hood_min
            5: hood_contrast
            6: hood_quantile
            7: hood_standard_deviation
            8: hood_median_absolute_deviation
            9: hood_z_score
            10: hood_z_score_mad
            11: hood_sigma_lee
        nodata (bool, default=False): If True, nodata values are considered
            in the convolution.
        nodata_value (float, default=-9999.9): The value representing nodata.

    Returns:
        numpy.ndarray: The convolved array.
    """
    x_adj = arr.shape[0] - 1
    y_adj = arr.shape[1] - 1
    z_adj = (arr.shape[2] - 1) // 2

    hood_size = len(offsets)

    result = np.zeros((arr.shape[0], arr.shape[1], 1), dtype="float32")

    for idx_x in prange(arr.shape[0]):
        for idx_y in range(arr.shape[1]):

            hood_values = np.zeros(hood_size, dtype="float32")

            hood_weights = np.zeros(hood_size, dtype="float32")
            weight_sum = np.array([0.0], dtype="float32")
            center_value = np.float32(arr[idx_x, idx_y, 0])
            normalise = False

            for idx_n in range(hood_size):
                offset_x = idx_x + offsets[idx_n][0]
                offset_y = idx_y + offsets[idx_n][1]
                offset_z = offsets[idx_n][2]

                outside = False

                if offset_z < -z_adj:
                    offset_z = -z_adj
                    outside = True
                elif offset_z > z_adj:
                    offset_z = z_adj
                    outside = True

                if offset_x < 0:
                    offset_x = 0
                    outside = True
                elif offset_x > x_adj:
                    offset_x = x_adj
                    outside = True

                if offset_y < 0:
                    offset_y = 0
                    outside = True
                elif offset_y > y_adj:
                    offset_y = y_adj
                    outside = True

                value = np.float32(arr[offset_x, offset_y, offset_z])

                if outside is True:
                    normalise = True
                    hood_weights[idx_n] = 0
                elif nodata and value == nodata_value:
                    normalise = True
                    hood_weights[idx_n] = 0
                else:
                    hood_values[idx_n] = value

                    weight = weights[idx_n]

                    hood_weights[idx_n] = weight
                    weight_sum[0] += weight

            if normalise and weight_sum[0] > 0.0:
                hood_weights = np.true_divide(hood_weights, weight_sum[0])

            if method == 1:
                result[idx_x, idx_y, 0] = hood_sum(hood_values, hood_weights)
            elif method == 2:
                result[idx_x, idx_y, 0] = hood_mode(hood_values, hood_weights)
            elif method == 3:
                result[idx_x, idx_y, 0] = hood_max(hood_values, hood_weights)
            elif method == 4:
                result[idx_x, idx_y, 0] = hood_min(hood_values, hood_weights)
            elif method == 5:
                result[idx_x, idx_y, 0] = hood_contrast(hood_values, hood_weights)
            elif method == 6:
                result[idx_x, idx_y, 0] = hood_quantile(hood_values, hood_weights, 0.5)
            elif method == 7:
                result[idx_x, idx_y, 0] = hood_standard_deviation(hood_values, hood_weights)
            elif method == 8:
                result[idx_x, idx_y, 0] = hood_median_absolute_deviation(hood_values, hood_weights)
            elif method == 9:
                result[idx_x, idx_y, 0] = hood_z_score(hood_values, center_value, hood_weights)
            elif method == 10:
                result[idx_x, idx_y, 0] = hood_z_score_mad(hood_values, center_value, hood_weights)
            elif method == 11:
                result[idx_x, idx_y, 0] = hood_sigma_lee(hood_values, hood_weights)
            else:
                result[idx_x, idx_y, 0] = nodata_value

    return result



""" Create patches from rasters, used for machine learnign applications. """

from typing import Union, List, Tuple, Optional, Callable
import numpy as np
from numba import prange, jit


def get_kernel_weights(
    tile_size: int = 64,
    edge_distance: int = 5,
    epsilon: float = 1e-7,
) -> np.ndarray:
    """
    Weight a kernel according to how close to an edge a given pixel is.

    Keyword Args:
        tile_size (int=64): The size of the square kernel.
        edge_distance (int=5): The distance from the edge to consider for weighting.
        epsilon (float=1e-7): A small value to prevent division by zero.

    Returns:
        np.ndarray: A 2D NumPy array of shape (tile_size, tile_size) with the kernel weights.
    """
    assert tile_size > 0, "Tile size must be greater than zero."
    assert edge_distance < tile_size // 2, "Edge distance must be less than half the tile size."
    assert edge_distance >= 0, "Edge distance must be greater than or equal to zero."

    arr = np.zeros((tile_size, tile_size), dtype="float32")
    max_dist = edge_distance * 2

    # Iterate through the kernel array
    for idx_y in range(0, arr.shape[0]):
        for idx_x in range(0, arr.shape[1]):

            # Calculate vertical distance to the closest edge
            val_y_top = max(edge_distance - idx_y, 0.0)
            val_y_bot = max((1 + edge_distance) - (tile_size - idx_y), 0.0)
            val_y = val_y_top + val_y_bot

            # Calculate horizontal distance to the closest edge
            val_x_lef = max(edge_distance - idx_x, 0.0)
            val_x_rig = max((1 + edge_distance) - (tile_size - idx_x), 0.0)
            val_x = val_x_lef + val_x_rig

            # Calculate the weight based on the distance to the closest edge
            val = (max_dist - abs(val_y + val_x)) / max_dist

            # Set a minimum weight to avoid division by zero
            if val <= 0.0:
                val = epsilon

            # Assign the calculated weight to the kernel array
            arr[idx_y, idx_x] = val

    return arr


@jit(nopython=True, parallel=True, nogil=True)
def merge_weighted_median(
    arr: np.ndarray,
    arr_weight: np.ndarray,
) -> np.ndarray:
    """
    Calculate the weighted median of a multi-dimensional array along the first axis.
    This is the order (number_of_overlaps, tile_size, tile_size, number_of_bands)
    
    Args:
        arr (np.ndarray): The input array.
        arr_weight (np.ndarray): The weight array with the same shape as the input array.
    
    Returns:
        np.ndarray: A 3D NumPy array of shape (arr.shape[1], arr.shape[2], arr.shape[3]) with the weighted medians.
    """
    ret_arr = np.empty((arr.shape[1], arr.shape[2], arr.shape[3]), dtype="float32")
    ret_arr[:] = np.nan

    # Iterate through the input array
    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            for idx_band in range(arr.shape[3]):

                # Flatten the input and weight arrays
                values = arr[:, idx_y, idx_x, idx_band].flatten()
                weights = arr_weight[:, idx_y, idx_x, 0].flatten()

                nan_mask = np.where(~np.isnan(values))[0]

                if len(nan_mask) == 0:
                    continue

                values = values[nan_mask]
                weights = weights[nan_mask]

                # Sort the values and weights based on the values
                sort_mask = np.argsort(values)
                sorted_data = values[sort_mask]
                sorted_weights = weights[sort_mask]

                # Calculate the cumulative sum of the sorted weights
                cumsum = np.cumsum(sorted_weights)

                # Normalize the cumulative sum to the range [0, 1]
                intersect = (cumsum - 0.5 * sorted_weights) / cumsum[-1]

                # Interpolate the weighted median and store it in the result array
                ret_arr[idx_y, idx_x, idx_band] = np.interp(0.5, intersect, sorted_data)

    return ret_arr


@jit(nopython=True, parallel=True, nogil=True)
def merge_weighted_average(
    arr: np.ndarray,
    arr_weight: np.ndarray,
) -> np.ndarray:
    """
    Calculate the weighted average of a multi-dimensional array along the last axis.
    
    Args:
        arr (np.ndarray): The input array.
        arr_weight (np.ndarray): The weight array with the same shape as the input array.
    
    Returns:
        np.ndarray: A 3D NumPy array of shape (arr.shape[0], arr.shape[1], 1) with the weighted averages.
    """
    ret_arr = np.empty((arr.shape[1], arr.shape[2], arr.shape[3]), dtype="float32")
    ret_arr[:] = np.nan

    # Iterate through the input array
    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            for idx_band in range(arr.shape[3]):

                # Flatten the input and weight arrays
                values = arr[:, idx_y, idx_x, idx_band].flatten()
                weights = arr_weight[:, idx_y, idx_x, 0].flatten()

                nan_mask = np.where(~np.isnan(values))[0]

                if len(nan_mask) == 0:
                    continue

                values = values[nan_mask]
                weights = weights[nan_mask]

                # Calculate the weighted sum and total weight
                weighted_sum = np.nansum(values * weights)
                total_weight = np.nansum(weights)

                # Calculate the weighted average and store it in the result array
                ret_arr[idx_y, idx_x, idx_band] = weighted_sum / total_weight

    return ret_arr


@jit(nopython=True, parallel=True, nogil=True)
def merge_weighted_mad(
    arr: np.ndarray,
    arr_weight: np.ndarray,
    mad_dist: float = 2.0,
) -> np.ndarray:
    """
    Merge an array of predictions using the MAD-merge methodology.
    
    Args:
        arr (np.ndarray): The input array.
        arr_weight (np.ndarray): The weight array with the same shape as the input array.
    Keyword Args:
        mad_dist (float=2.0): The MAD distance.
    
    Returns:
        np.ndarray: A 3D NumPy array of shape (arr.shape[0], arr.shape[1], 1) with the MAD-merged values.
    """
    ret_arr = np.empty((arr.shape[1], arr.shape[2], arr.shape[3]), dtype="float32")
    ret_arr[:] = np.nan

    # Iterate through the input array
    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            for idx_band in range(arr.shape[3]):

                # Flatten the input and weight arrays
                values = arr[:, idx_y, idx_x, idx_band].flatten()
                weights = arr_weight[:, idx_y, idx_x, 0].flatten()

                nan_mask = np.where(~np.isnan(values))[0]

                if len(nan_mask) == 0:
                    continue

                values = values[nan_mask]
                weights = weights[nan_mask]

                # Sort the values and weights based on the values
                sort_mask = np.argsort(values)
                sorted_data = values[sort_mask]
                sorted_weights = weights[sort_mask]

                # Calculate the cumulative sum of the sorted weights and normalize to the range [0, 1]
                cumsum = np.cumsum(sorted_weights)
                intersect = (cumsum - 0.5 * sorted_weights) / cumsum[-1]

                # Interpolate the median
                median = np.interp(0.5, intersect, sorted_data)

                # Calculate the median absolute deviation (MAD)
                mad = np.median(np.abs(median - values))

                # If MAD is zero, store the median in the result array and continue
                if mad == 0.0:
                    ret_arr[idx_y, idx_x, 0] = median
                    continue

                # Calculate the new weights based on the MAD
                new_weights = np.zeros_like(sorted_weights)
                for idx_z in range(sorted_data.shape[0]):
                    new_weights[idx_z] = 1.0 - (np.minimum(np.abs(sorted_data[idx_z] - median) / (mad * mad_dist), 1))

                if np.sum(new_weights) == 0.0:
                    ret_arr[idx_y, idx_x, 0] = median
                    continue

                # Calculate the cumulative sum of the new weights and normalize to the range [0, 1]
                cumsum = np.cumsum(new_weights)
                intersect = (cumsum - 0.5 * new_weights) / cumsum[-1]

                # Interpolate the MAD-merged value and store it in the result array
                ret_arr[idx_y, idx_x, idx_band] = np.interp(0.5, intersect, sorted_data)

    return ret_arr


@jit(nopython=True, nogil=True)
def unique_values(arr: np.ndarray) -> np.ndarray:
    """
    Find the unique values in a 1D NumPy array.
    
    Args:
        arr (np.ndarray): The input array.
    
    Returns:
        np.ndarray: A 1D NumPy array with the unique values.
    """
    unique = np.empty(arr.size, dtype=arr.dtype)
    unique_count = 0
    for i in range(arr.shape[0]):
        if arr[i] not in unique[:unique_count]:
            unique[unique_count] = arr[i]
            unique_count += 1

    return unique[:unique_count]


@jit(nopython=True, parallel=True, nogil=True)
def merge_weighted_mode(
    arr: np.ndarray,
    arr_weight: np.ndarray,
) -> np.ndarray:
    """
    Calculate the weighted mode of a multi-dimensional array along the last axis.
    
    Args:
        arr (np.ndarray): The input array.
        arr_weight (np.ndarray): The weight array with the same shape as the input array.
    
    Returns:
        np.ndarray: A 3D NumPy array of shape (arr.shape[0], arr.shape[1], 1) with the weighted modes.
    """
    ret_arr = np.empty((arr.shape[1], arr.shape[2], arr.shape[3]), dtype=arr.dtype)
    ret_arr[:] = np.nan

    # Iterate through the input array
    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            for idx_band in range(arr.shape[3]):

                # Flatten the input and weight arrays
                values = arr[:, idx_y, idx_x, idx_band].flatten()
                weights = arr_weight[:, idx_y, idx_x, 0].flatten()

                nan_mask = np.where(~np.isnan(values))[0]

                if len(nan_mask) == 0:
                    continue

                values = values[nan_mask]
                weights = weights[nan_mask]

                # Get unique values and their weighted counts
                unique_vals = unique_values(values)
                weighted_counts = np.zeros(unique_vals.shape[0])

                # Calculate the weighted sum for each unique value
                for i in range(unique_vals.shape[0]):
                    idxs = np.where(values == unique_vals[i])
                    weighted_counts[i] = np.sum(weights[idxs])

                # Get the index of the maximum weighted sum
                mode_idx = np.argmax(weighted_counts)

                # Store the weighted mode in the result array
                ret_arr[idx_y, idx_x, idx_band] = unique_vals[mode_idx]

    return ret_arr


def calculate_offset_single(
    tile_size: int,
    num_offsets: int,
) -> List[int]:
    """
    Calculate a list of offset values for a given tile size and number of offsets.

    Args:
        tile_size (int): The size of each tile.
        num_offsets (int): The desired number of offsets to be calculated.

    Returns:
        List[int]: A list of calculated offset values.
    """
    assert num_offsets >= 0, "Number of offsets must be greater than or equal to 0."
    assert tile_size > 0, "Tile size must be greater than 0."

    # Initialize the list of offsets with the first value, 0
    offsets = [0]
    previous = None
    for _ in range(num_offsets):

        # Calculate the step size based on previous step
        if previous is None:
            step = tile_size // 2
            previous = step
        else:
            step = previous // 2
            previous = step

        # Break the loop if the step size is 0
        if step == 0:
            break

        # Calculate and add offset values based on the step size
        for j in range(1, num_offsets + 1):
            val = j * step
            if (
                val not in offsets
                and val < tile_size
                and val > 0
                and len(offsets) < num_offsets + 1
            ):
                offsets.append(j * step)

    return offsets

def get_offsets(
    tile_size: int,
    offsets_y: int,
    offsets_x: int,
) -> List[Tuple[int, int]]:
    """
    Generate a list of offset pairs for a given tile size and number of offsets in x and y dimensions.

    Args:
        tile_size (int): The size of each tile.
        offsets_y (int): The desired number of offsets to be calculated in the y dimension.
        offsets_x (int): The desired number of offsets to be calculated in the x dimension.

    Returns:
        List[Tuple[int, int]]: A list of tuples containing offset pairs for y and x dimensions.
        order is (y, x)
    """
    assert offsets_y >= 0, "Number of offsets in y dimension must be >= 0"
    assert offsets_x >= 0, "Number of offsets in x dimension must be >= 0"
    assert tile_size > 0, "Tile size must be > 0"

    offsets_y = calculate_offset_single(tile_size, offsets_y)
    offsets_x = calculate_offset_single(tile_size, offsets_x)

    offsets = []
    for y in offsets_y:
        for x in offsets_x:
            offsets.append((y, x))

    return offsets


def borders_are_necessary(
    arr: np.ndarray,
    tile_size: int,
    offset: List[int],
) -> Tuple[bool, bool]:
    """
    Checks if borders are necessary for the given array.
    Width and height are returned as a tuple.

    Args:
        arr (np.ndarray): The array to be checked.
        tile_size (int): The size of each tile.
        offset (list): The offset to be used.

    Returns:
        tuple: A tuple containing of borders are needed in (height, width) dims.
    """
    if arr.ndim == 2:
        height, width = arr.shape
    else:
        height, width, _ = arr.shape

    if (height - offset[0]) % tile_size == 0:
        height_border = False
    else:
        height_border = True

    if (width - offset[1]) % tile_size == 0:
        width_border = False
    else:
        width_border = True

    return height_border, width_border


def borders_are_necessary_list(
    arr: np.ndarray,
    tile_size: int,
    offsets: List[List[int]],
) -> Tuple[bool, bool]:
    """
    Checks if borders are necessary for the given array.
    Width and height are returned as a tuple.

    Args:
        arr (np.ndarray): The array to be checked.
        tile_size (int): The size of each tile.
        offsets (list): The offset to be used.

    Returns:
        tuple: A tuple containing of borders are needed in (height, width) dims.
    """
    height_border = True
    width_border = True

    for offset in offsets:
        offset_height_border, offset_width_border = borders_are_necessary(
            arr, tile_size, offset
        )
        if not offset_height_border:
            height_border = False

        if not offset_width_border:
            width_border = False

        if not height_border and not width_border:
            break

    return height_border, width_border


def array_to_patches_single(
    arr: np.ndarray,
    tile_size: int,
    offset: Optional[Union[List[int], Tuple[int, int]]] = None,
) -> np.ndarray:
    """
    Generate patches from an array.

    Args:
        arr (np.ndarray): A numpy array to be divided into patches.
        tile_size (int): The size of each tile/patch, e.g., 64 for 64x64 tiles.
        offset (list/tuple/None=None): The x and y offset values for the input
            array. If not provided, defaults to [0, 0].

    Returns:
        nd.array: A numpy array containing the patches.
    """
    assert arr.ndim in [2, 3], "Array must be 2D or 3D"
    assert tile_size > 0, "Tile size must be greater than 0"
    assert offset is None or len(offset) == 2, "Offset must be a list or tuple of length 2"

    # Set default offset to [0, 0] if not provided
    if offset is None:
        offset = [0, 0]

    # Calculate the number of patches in the y and x dimensions
    patches_y = (arr.shape[0] - offset[0]) // tile_size
    patches_x = (arr.shape[1] - offset[1]) // tile_size

    # Calculate cut dimensions for the y and x dimensions
    cut_y = -((arr.shape[0] - offset[0]) % tile_size)
    cut_x = -((arr.shape[1] - offset[1]) % tile_size)

    # Set cut dimensions to None if they are 0
    cut_y = None if cut_y == 0 else cut_y
    cut_x = None if cut_x == 0 else cut_x

    # Reshape the array to separate the patches
    reshaped = arr[offset[0]:cut_y, offset[1]:cut_x].reshape(
        patches_y,
        tile_size,
        patches_x,
        tile_size,
        arr.shape[2],
    )

    # Swap axes to rearrange patches in the correct order
    swaped = reshaped.swapaxes(1, 2)

    # Combine the patches into a single array
    blocks = swaped.reshape(-1, tile_size, tile_size, arr.shape[2])

    return blocks


def patches_to_array_single(
    patches: np.ndarray,
    shape: Union[List, Tuple],
    tile_size: int,
    offset: Optional[Union[List, Tuple]] = None,
    background_value: Optional[Union[int, float]] = None,
) -> np.ndarray:
    """
    Reconstitute an array from patches.

    Given an array of patches, this function stitches them back together
    to form the original array of the specified shape.

    Args:
        patches (np.ndarray): A numpy array containing the patches to be
            stitched together.
        shape (list/tuple): The desired shape of the output array.
        tile_size (int): The size of each tile/patch Eg. 64 for 64x64 tiles.

    Keyword Args:
        offset (list/tuple/None=None): The x and y offset values for the
            target array. If not provided, defaults to [0, 0].

    Returns:
        np.ndarray: A numpy array with the original shape, formed by stitching
            together the provided patches.
    """
    assert len(shape) in [2, 3], "Shape must be a tuple or list of length 2 or 3"
    assert len(patches.shape) == 4, "Patches must be a 4D array"
    assert patches.shape[1] == tile_size, "Patches must be of size tile_size"
    assert patches.shape[2] == tile_size, "Patches must be of size tile_size"
    assert offset is None or len(offset) == 2, "Offset must be a tuple or list of length 2"

    # Set default offset to [0, 0] if not provided
    if offset is None:
        offset = [0, 0]

    # Create an empty target array of the specified shape
    if background_value is None:
        target = np.full(shape, np.nan, dtype=patches.dtype)
    else:
        target = np.full(shape, background_value, dtype=patches.dtype)

    # Calculate target dimensions
    target_y = ((shape[0] - offset[0]) // tile_size) * tile_size
    target_x = ((shape[1] - offset[1]) // tile_size) * tile_size

    # Calculate cut dimensions
    cut_y = -((shape[0] - offset[0]) % tile_size)
    cut_x = -((shape[1] - offset[1]) % tile_size)

    # Set cut dimensions to None if they are 0
    cut_y = None if cut_y == 0 else cut_y
    cut_x = None if cut_x == 0 else cut_x

    # Calculate the number of tiles in the y and x dimensions
    num_tiles_y = target_y // tile_size
    num_tiles_x = target_x // tile_size

    # Reshape the patches for stitching
    reshape = patches.reshape(
        num_tiles_y,
        num_tiles_x,
        tile_size,
        tile_size,
        patches.shape[3],
        1,
    )

    # Swap axes to rearrange patches in the correct order for stitching
    swap = reshape.swapaxes(1, 2)

    # Combine the patches into a single array
    destination = swap.reshape(
        num_tiles_y * tile_size,
        num_tiles_x * tile_size,
        patches.shape[3],
    )

    # Assign the combined patches to the target array
    target[offset[0]:cut_y, offset[1]:cut_x] = destination

    return target


def patches_to_weights(
    patches: np.ndarray,
    edge_distance: int,
) -> np.ndarray:
    """ Calculate the weights for each patch based on the distance to the edge. """
    assert len(patches.shape) == 4, "Patches must be a 4D array"
    assert patches.shape[1] == patches.shape[2], "Patches must be square"

    # Calculate the distance to the edge for each patch
    weights = get_kernel_weights(patches.shape[1], edge_distance)

    # Expand the weights to match the number of patches
    weights = np.repeat(weights[np.newaxis, ...], patches.shape[0], axis=0)[..., np.newaxis]

    return weights

def array_to_patches(
    arr: np.ndarray,
    tile_size: int,
    offsets_y: int = 3,
    offsets_x: int = 3,
    border_check: bool = True,
) -> np.ndarray:
    """
    Generate patches from an array based on the specified parameters.

    Args:
        arr (np.ndarray): A numpy array to be divided into patches.
        tile_size (int): The size of each tile/patch, e.g., 64 for 64x64 tiles.
        offsets_y (int=3): The desired number of offsets to be calculated in the y dimension.
        offsets_x (int=3): The desired number of offsets to be calculated in the x dimension.
        border_check (bool=True): Whether or not to include border patches.

    Returns:
        np.ndarray: The concatenate patches along axis 0. In the order (patches, y, x, channels)
    """
    # Get the list of offsets for both x and y dimensions
    offsets = get_offsets(tile_size, offsets_y, offsets_x)

    if border_check:
        borders_y, borders_x = borders_are_necessary_list(arr, tile_size, offsets)

        if borders_y:
            offsets.append([arr.shape[0] - tile_size, 0])
        if borders_x:
            offsets.append([0, arr.shape[1] - tile_size])

    # Initialize an empty list to store the generated patches
    patches = []

    # Iterate through the offsets and generate patches for each offset
    for offset in offsets:
        patches.append(
            array_to_patches_single(arr, tile_size, offset),
        )

    patches = np.concatenate(patches, axis=0)

    return patches


def predict_array(
    arr: np.ndarray,
    callback: Callable[[np.ndarray], np.ndarray],
    tile_size=64,
    offsets_y=1,
    offsets_x=1,
    border_check=True,
    merge_method="median",
    edge_weighted=True,
    edge_distance=3,
) -> np.ndarray:
    """
    Generate patches from an array. Also outputs the offsets and the shapes of the offsets. Only
    suppors the prediction of single values in the rasters/arrays. 
    
    Args:
        arr (np.ndarray): A numpy array to be divided into patches.
        callback (function): The callback function to be used for prediction. The callback function
            must take a numpy array as input and return a numpy array as output.
    
    Keyword Args:
        tile_size (int=64): The size of each tile/patch, e.g., 64 for 64x64 tiles.
        offsets_y (int=1): The desired number of offsets to be calculated in the y dimension.
        offsets_x (int=1): The desired number of offsets to be calculated in the x dimension.
        border_check (bool=True): Whether or not to include border patches.
        merge_method (str="median"): The method to use for merging the patches. Valid methods
        are ['mad', 'median', 'mean', 'mode']
        edge_weighted (bool=True): Whether or not to weight the edges patches of patches less
            than the central parts.
        edge_distance (int=3): The distance from the edge to be weighted less. Usually good to
            adjust this to your maximum convolution kernel size.

    Returns:
        np.ndarray: The predicted array.
    """
    assert merge_method in ["mad", "median", "mean", "mode"], "Invalid merge method"
    assert len(arr.shape) == 3, "Array must be 3D"
    assert tile_size < arr.shape[0], "Tile size must be smaller than the array size"
    assert tile_size < arr.shape[1], "Tile size must be smaller than the array size"

    # Get the list of offsets for both x and y dimensions
    offsets = get_offsets(tile_size, offsets_y, offsets_x)

    if border_check:
        borders_y, borders_x = borders_are_necessary_list(arr, tile_size, offsets)

        if borders_y:
            offsets.append([arr.shape[0] - tile_size, 0])
        if borders_x:
            offsets.append([0, arr.shape[1] - tile_size])

    # Test output dimensions of prediction
    test_patch = arr[np.newaxis, :tile_size, :tile_size, :]
    test_prediction = callback(test_patch)
    test_shape = test_prediction.shape

    # Initialize an empty list to store the generated patches
    predictions = np.zeros((len(offsets), arr.shape[0], arr.shape[1], test_shape[-1]), dtype=arr.dtype)
    predictions_weights = np.zeros((len(offsets), arr.shape[0], arr.shape[1], 1), dtype=np.float32)

    # Iterate through the offsets and generate patches for each offset
    for idx, offset in enumerate(offsets):
        patches = array_to_patches_single(arr, tile_size, offset)

        prediction = callback(patches)

        if edge_weighted:
            weights = patches_to_weights(patches, edge_distance)
        else:
            weights = np.ones((tile_size, tile_size, 1), dtype=np.float32)
            weights = np.repeat(weights[np.newaxis, ...], patches.shape[0], axis=0)

        predictions[idx, :, :, :] = patches_to_array_single(prediction, (arr.shape[0], arr.shape[1], test_shape[-1]), tile_size, offset)
        predictions_weights[idx, :, :, :] = patches_to_array_single(weights, (arr.shape[0], arr.shape[1], 1), tile_size, offset, 0.0)

    # Merge the predictions
    if merge_method == "mad":
        predictions = merge_weighted_mad(predictions, predictions_weights)
    elif merge_method == "median":
        predictions = merge_weighted_median(predictions, predictions_weights)
    elif merge_method == "mean" or merge_method == "average":
        predictions = merge_weighted_average(predictions, predictions_weights)
    elif merge_method == "mode":
        predictions = merge_weighted_mode(predictions, predictions_weights)

    return predictions



"""
### Clip vectors to other geometries ###

Clip vector files with other geometries. Can come from rasters or vectors.
"""

# Standard library
import sys; sys.path.append("../../")

# External
from osgeo import ogr, gdal, osr

# Internal
from buteo.utils import core_utils, gdal_utils
from buteo.raster import core_raster
from buteo.vector import core_vector
from buteo.vector.reproject import _reproject_vector



def _clip_vector(
    vector,
    clip_geom,
    out_path=None,
    *,
    to_extent=False,
    target_projection=None,
    preserve_fid=True,
    promote_to_multi=True,
):
    """ Internal. """

    input_path = gdal_utils.get_path_from_dataset(vector)

    if out_path is None:
        out_path = gdal_utils.create_memory_path(
            input_path,
            prefix="",
            suffix="_clip",
            add_uuid=True,
        )

    assert core_utils.is_valid_output_path(out_path), "Invalid output path"

    options = []

    clear_memory = False
    geometry_to_clip = None
    if gdal_utils.is_vector(clip_geom):
        if to_extent:
            extent = core_vector._vector_to_metadata(clip_geom)["get_bbox_vector"]() # pylint: disable=not-callable
            geometry_to_clip = extent
            clear_memory = True
        else:
            geometry_to_clip = core_vector._open_vector(clip_geom)
    elif gdal_utils.is_raster(clip_geom):
        extent = core_raster._raster_to_metadata(clip_geom)["get_bbox_vector"]() # pylint: disable=not-callable
        geometry_to_clip = extent
        clear_memory = True
    else:
        raise ValueError(f"Invalid input in clip_geom, unable to parse: {clip_geom}")

    clip_vector_path = gdal_utils.get_path_from_dataset(geometry_to_clip)
    clip_vector_reprojected = _reproject_vector(clip_vector_path, vector)

    if clear_memory:
        gdal_utils.delete_if_in_memory(clip_vector_path)

    x_min, x_max, y_min, y_max = core_vector._vector_to_metadata(clip_vector_reprojected)["extent"]

    options.append(f"-spat {x_min} {y_min} {x_max} {y_max}")

    options.append(f'-clipsrc "{clip_vector_reprojected}"')

    if promote_to_multi:
        options.append("-nlt PROMOTE_TO_MULTI")

    if preserve_fid:
        options.append("-preserve_fid")
    else:
        options.append("-unsetFid")

    if target_projection is not None:
        wkt = gdal_utils.parse_projection(target_projection, return_wkt=True).replace(" ", "\\")

        options.append(f'-t_srs "{wkt}"')

    # dst  # src
    success = gdal.VectorTranslate(
        out_path,
        input_path,
        format=gdal_utils.path_to_driver_vector(out_path),
        options=" ".join(options),
    )

    gdal_utils.delete_if_in_memory(clip_vector_reprojected)

    if success != 0:
        return out_path
    else:
        raise Exception("Error while clipping geometry.")


def clip_vector(
    vector,
    clip_geom,
    out_path=None,
    *,
    to_extent=False,
    target_projection=None,
    preserve_fid=True,
    prefix="",
    suffix="",
    add_uuid=False,
    allow_lists=True,
    overwrite=True,
    promote_to_multi=True,
):
    """
    Clips a vector to a geometry.

    ## Args:
    `vector` (_str_/_ogr.DataSource_/_list_): Vector(s) to clip. </br>
    `clip_geom` (_str_/_ogr.Geometry_): Vector to clip with. </br>

    ## Kwargs:
    `out_path` (_str_/_None_): Output path. If None, memory vectors are created. (Default: **None**) </br>
    `to_extent` (_bool_): Clip to extent. (Default: **False**) </br>
    `target_projection` (_str_/_ogr.DataSource_/_gdal.Dataset_/_osr.SpatialReference_/_int_/_None_): Target projection. (Default: **None**) </br>
    `preserve_fid` (_bool_): Preserve fid. (Default: **True**) </br>
    `prefix` (_str_): Prefix to add to the output path. (Default: **""**) </br>
    `suffix` (_str_): Suffix to add to the output path. (Default: **""**) </br>
    `add_uuid` (_bool_): Add UUID to the output path. (Default: **False**) </br>
    `allow_lists` (_bool_): Allow lists of vectors as input. (Default: **True**) </br>
    `overwrite` (_bool_): Overwrite output if it already exists. (Default: **True**) </br>
    `promote_to_multi` (_bool_): Should POLYGON by promoted to MULTIPOLYGON.. (Default: **True**) </br>

    ## Returns:
    (_str_/_list_): Output path(s) of clipped vector(s).
    """
    core_utils.type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")
    core_utils.type_check(clip_geom, [ogr.DataSource, gdal.Dataset, str, list, tuple], "clip_geom")
    core_utils.type_check(out_path, [str, None], "out_path")
    core_utils.type_check(to_extent, [bool], "to_extent")
    core_utils.type_check(target_projection, [str, ogr.DataSource, gdal.Dataset, osr.SpatialReference, int, None], "target_projection")
    core_utils.type_check(preserve_fid, [bool], "preserve_fid")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "suffix")
    core_utils.type_check(add_uuid, [bool], "add_uuid")
    core_utils.type_check(allow_lists, [bool], "allow_lists")

    if not allow_lists and isinstance(vector, (list, tuple)):
        raise ValueError("Lists are not allowed for vector.")

    vector_list = core_utils.ensure_list(vector)

    assert gdal_utils.is_vector_list(vector_list), f"Invalid vector in list: {vector_list}"

    path_list = gdal_utils.create_output_path_list(
        vector_list,
        out_path=out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        overwrite=overwrite,
    )

    output = []
    for index, in_vector in enumerate(vector_list):
        output.append(
            _clip_vector(
                in_vector,
                clip_geom,
                out_path=path_list[index],
                to_extent=to_extent,
                target_projection=target_projection,
                preserve_fid=preserve_fid,
                promote_to_multi=promote_to_multi,
            )
        )

    if isinstance(vector, list):
        return output

    return output[0]



"""
### Generic utility functions ###

Functions that make interacting with the toolbox easier.
"""
# Internal
from typing import Tuple

# External
import numpy as np
from numba import jit, prange



@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_latitude(lat):
    """ Latitude goes from -90 to 90 """
    lat_adj = lat + 90.0
    lat_max = 180

    encoded_sin = ((np.sin(2 * np.pi * (lat_adj / lat_max)) + 1)) / 2.0
    encoded_cos = ((np.cos(2 * np.pi * (lat_adj / lat_max)) + 1)) / 2.0

    return np.array([encoded_sin, encoded_cos], dtype=np.float32)

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_longitude(lng):
    """ Longitude goes from -180 to 180 """
    lng_adj = lng + 180.0
    lng_max = 360

    encoded_sin = ((np.sin(2 * np.pi * (lng_adj / lng_max)) + 1)) / 2.0
    encoded_cos = ((np.cos(2 * np.pi * (lng_adj / lng_max)) + 1)) / 2.0

    return np.array([encoded_sin, encoded_cos], dtype=np.float32)

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_latlng(latlng):
    """
    Encode latitude and longitude values to be used as input to the model.
    """
    lat = latlng[0]
    lng = latlng[1]

    encoded_lat = encode_latitude(lat)
    encoded_lng = encode_longitude(lng)

    return np.concatenate((encoded_lat, encoded_lng)).astype(np.float32)

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_latlngs(latlngs):
    """ Encode multiple latitude and longitude values. """
    if latlngs.ndim == 1:
        encoded_latlngs = np.apply_along_axis(encode_latlng, 0, latlngs)
    elif latlngs.ndim == 2:
        encoded_latlngs = np.apply_along_axis(encode_latlng, 1, latlngs)
    elif latlngs.ndim == 3:
        rows = latlngs.shape[0]
        cols = latlngs.shape[1]

        output_shape = (rows, cols, 4)
        encoded_latlngs = np.zeros(output_shape, dtype=np.float32)

        for i in prange(rows):
            for j in range(cols):
                latlng = latlngs[i, j]
                encoded_latlngs[i, j] = encode_latlng(latlng)
    else:
        raise ValueError(
            f"The input array must have 1, 2 or 3 dimensions, not {latlngs.ndim}"
        )

    return encoded_latlngs


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_latitude(encoded_sin, encoded_cos):
    """
    Decode encoded latitude values to the original latitude value.
    """
    lat_max = 180
    lat_max_half = lat_max / 2.0

    # Calculate the sin and cos values from the encoded values
    sin_val = (2 * encoded_sin) - 1
    cos_val = (2 * encoded_cos) - 1

    # Calculate the latitude adjustment
    lat_adj = np.arctan2(sin_val, cos_val)

    # Convert the adjusted latitude to the original latitude value
    sign = np.sign(lat_adj)
    sign_adj = np.where(sign == 0, 1, sign) * lat_max_half

    lat = ((lat_adj / (2 * np.pi)) * lat_max) - sign_adj
    lat = np.where(lat == -lat_max_half, lat_max_half, lat)

    return lat

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_longitude(encoded_sin, encoded_cos):
    """
    Decode encoded longitude values to the original longitude value.
    """
    lng_max = 360
    lng_max_half = lng_max / 2.0

    # Calculate the sin and cos values from the encoded values
    sin_val = (2 * encoded_sin) - 1
    cos_val = (2 * encoded_cos) - 1

    # Calculate the longitude adjustment
    lng_adj = np.arctan2(sin_val, cos_val)

    # Convert the adjusted longitude to the original longitude value
    sign = np.sign(lng_adj)
    sign_adj = np.where(sign == 0, 1, sign) * lng_max_half

    lng = ((lng_adj / (2 * np.pi)) * lng_max) - sign_adj

    lng = np.where(lng == -lng_max_half, lng_max_half, lng)

    return lng


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_latlng(encoded_latlng):
    """
    Decode encoded latitude and longitude values to the original values.
    """
    lat = decode_latitude(encoded_latlng[0], encoded_latlng[1])
    lng = decode_longitude(encoded_latlng[2], encoded_latlng[3])

    return np.array([lat, lng], dtype=np.float32)

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_latlngs(encoded_latlngs):
    """ Decode multiple latitude and longitude values. """
    latlngs = np.apply_along_axis(decode_latlng, 1, encoded_latlngs)
    return latlngs


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_width(lng, lng_max):
    """ Longitude goes from -180 to 180 """

    encoded_sin = ((np.sin(2 * np.pi * (lng / lng_max)) + 1)) / 2.0
    encoded_cos = ((np.cos(2 * np.pi * (lng / lng_max)) + 1)) / 2.0

    return np.array([encoded_sin, encoded_cos], dtype=np.float32)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_arr_position(arr):
    """ Fast encoding of coordinates where the width is cyclical. """
    result = np.zeros((arr.shape[0], arr.shape[1], 3), dtype=np.float32)

    col_end = arr.shape[0] - 1
    row_end = arr.shape[1] - 1

    col_range = np.arange(0, arr.shape[0]).astype(np.float32)
    row_range = np.arange(0, arr.shape[1]).astype(np.float32)

    col_encoded = np.zeros((col_range.shape[0], 1), dtype=np.float32)
    row_encoded = np.zeros((row_range.shape[0], 2), dtype=np.float32)

    for col in prange(col_range.shape[0]):
        col_encoded[col, :] = col_range[col] / col_end

    for row in prange(row_range.shape[0]):
        row_encoded[row, :] = encode_width(row_range[row], row_end)

    for col in prange(arr.shape[0]):
        for row in range(arr.shape[1]):
            result[col, row, 0] = row_encoded[row, 0]
            result[col, row, 1] = row_encoded[row, 1]
            result[col, row, 2] = col_encoded[col, 0]

    return result

def channel_first_to_last(arr):
    """ Converts a numpy array from channel first to channel last format. """
    if arr.ndim != 3:
        raise ValueError("Input array should be 3-dimensional with shape (channels, height, width)")

    # Swap the axes to change from channel first to channel last format
    arr = np.transpose(arr, (1, 2, 0))

    return arr

def channel_last_to_first(arr):
    """ Converts a numpy array from channel last to channel first format. """
    if arr.ndim != 3:
        raise ValueError("Input array should be 3-dimensional with shape (height, width, channels)")

    # Swap the axes to change from channel last to channel first format
    arr = np.transpose(arr, (2, 0, 1))

    return arr

def scale_to_range(arr, min_val, max_val):
    """ Scales the values in the input array to the specified range. """

    # Scale the values in the array to the specified range
    arr = (arr - arr.min()) / (arr.max() - arr.min())
    arr = (max_val - min_val) * arr + min_val

    return arr

@jit(nopython=True)
def create_grid(range_rows, range_cols):
    """ Create a grid of rows and columns """
    rows_grid = np.zeros((len(range_rows), len(range_cols)), dtype=np.int64)
    cols_grid = np.zeros((len(range_rows), len(range_cols)), dtype=np.int64)

    for i in range(len(range_rows)):
        for j in range(len(range_cols)):
            cols_grid[i, j] = range_rows[j]
            rows_grid[i, j] = range_cols[i]

    return rows_grid, cols_grid


def split_into_offsets(shape, offsets_x=2, offsets_y=2, overlap_x=0, overlap_y=0):
    """ Split a shape into offsets. Usually used for splitting an image into offsets to reduce RAM needed. """
    height = shape[0]
    width = shape[1]

    x_remainder = width % offsets_x
    y_remainder = height % offsets_y

    x_offsets = [0]
    x_sizes = []
    for _ in range(offsets_x - 1):
        x_offsets.append(x_offsets[-1] + (width // offsets_x) - overlap_x)
    x_offsets[-1] -= x_remainder

    for idx, _ in enumerate(x_offsets):
        if idx == len(x_offsets) - 1:
            x_sizes.append(width - x_offsets[idx])
        elif idx == 0:
            x_sizes.append(x_offsets[1] + overlap_x)
        else:
            x_sizes.append(x_offsets[idx + 1] - x_offsets[idx] + overlap_x)

    y_offsets = [0]
    y_sizes = []
    for _ in range(offsets_y - 1):
        y_offsets.append(y_offsets[-1] + (height // offsets_y) - overlap_y)
    y_offsets[-1] -= y_remainder

    for idx, _ in enumerate(y_offsets):
        if idx == len(y_offsets) - 1:
            y_sizes.append(height - y_offsets[idx])
        elif idx == 0:
            y_sizes.append(y_offsets[1] + overlap_y)
        else:
            y_sizes.append(y_offsets[idx + 1] - y_offsets[idx] + overlap_y)

    offsets = []

    for idx_col, _ in enumerate(y_offsets):
        for idx_row, _ in enumerate(x_offsets):
            offsets.append([
                x_offsets[idx_row],
                y_offsets[idx_col],
                x_sizes[idx_row],
                y_sizes[idx_col],
            ])

    return offsets


@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def calculate_pixel_distances(array, target=1, maximum_distance=None, pixel_width=1, pixel_height=1):
    """ Calculate the distance from each pixel to the nearest target pixel. """
    binary_array = np.sum(array == target, axis=2, dtype=np.uint8)

    if maximum_distance is None:
        maximum_distance = np.sqrt(binary_array.shape[0] ** 2 + binary_array.shape[1] ** 2)

    radius_cols = int(np.ceil(maximum_distance / pixel_height))
    radius_rows = int(np.ceil(maximum_distance / pixel_width))

    kernel_cols = radius_cols * 2
    kernel_rows = radius_rows * 2

    if kernel_cols % 2 == 0:
        kernel_cols += 1

    if kernel_rows % 2 == 0:
        kernel_rows += 1

    middle_cols = int(np.floor(kernel_cols / 2))
    middle_rows = int(np.floor(kernel_rows / 2))

    range_cols = np.arange(-middle_cols, middle_cols + 1)
    range_rows = np.arange(-middle_rows, middle_rows + 1)

    cols_grid, rows_grid = create_grid(range_rows, range_cols)
    coord_grid = np.empty((cols_grid.size, 2), dtype=np.int64)
    coord_grid[:, 0] = cols_grid.flatten()
    coord_grid[:, 1] = rows_grid.flatten()

    coord_grid_projected = np.empty_like(coord_grid, dtype=np.float32)
    coord_grid_projected[:, 0] = coord_grid[:, 0] * pixel_height
    coord_grid_projected[:, 1] = coord_grid[:, 1] * pixel_width

    coord_grid_values = np.sqrt((coord_grid_projected[:, 0] ** 2) + (coord_grid_projected[:, 1] ** 2))

    selected_range = np.arange(coord_grid.shape[0])
    selected_range = selected_range[np.argsort(coord_grid_values)][1:]
    selected_range = selected_range[coord_grid_values[selected_range] <= maximum_distance]

    coord_grid = coord_grid[selected_range]
    coord_grid_values = coord_grid_values[selected_range]

    distances = np.full_like(binary_array, maximum_distance, dtype=np.float32)
    for col in prange(binary_array.shape[0]):
        for row in range(binary_array.shape[1]):
            if binary_array[col, row] == target:
                distances[col, row] = 0
            else:
                for idx, (col_adj, row_adj) in enumerate(coord_grid):
                    if (col + col_adj) >= 0 and (col + col_adj) < binary_array.shape[0] and \
                        (row + row_adj) >= 0 and (row + row_adj) < binary_array.shape[1] and \
                        binary_array[col + col_adj, row + row_adj] == target:

                        distances[col, row] = coord_grid_values[idx]
                        break

    return np.expand_dims(distances, axis=2)

@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def fill_nodata_with_nearest_average(array, nodata_value, mask=None, max_iterations=None, channel=0):
    """ Calculate the distance from each pixel to the nearest target pixel. """
    kernel_size = 3

    range_rows = np.arange(-(kernel_size // 2), (kernel_size // 2) + 1)
    range_cols = np.arange(-(kernel_size // 2), (kernel_size // 2) + 1)

    cols_grid, rows_grid = create_grid(range_rows, range_cols)
    coord_grid = np.empty((cols_grid.size, 2), dtype=np.int64)
    coord_grid[:, 0] = cols_grid.flatten()
    coord_grid[:, 1] = rows_grid.flatten()

    coord_grid_values = np.sqrt((coord_grid[:, 0] ** 2) + (coord_grid[:, 1] ** 2))
    coord_grid_values_sort = np.argsort(coord_grid_values)[1:]
    coord_grid_values = coord_grid_values[coord_grid_values_sort]

    coord_grid = coord_grid[coord_grid_values_sort]

    weights = 1 / coord_grid_values
    weights = weights / np.sum(weights)
    weights = weights.astype(np.float32)

    main_filled = np.copy(array)

    if mask is None:
        mask = np.ones_like(main_filled, dtype=np.uint8)
    else:
        mask = (mask == 1).astype(np.uint8)

    main_filled = main_filled[:, :, channel]
    mask = mask[:, :, channel]

    nodata_value = np.array(nodata_value, dtype=array.dtype)
    uint8_1 = np.array(1, dtype=np.uint8)

    iterations = 0
    while True:
        local_filled = np.copy(main_filled)
        for row in prange(main_filled.shape[0]):
            for col in prange(main_filled.shape[1]):
                if main_filled[row, col] != nodata_value:
                    continue
                if mask[row, col] != uint8_1:
                    continue

                count = 0
                weights_sum = 0.0
                value_sum = 0.0

                for idx, (col_adj, row_adj) in enumerate(coord_grid):
                    if (row + row_adj) >= 0 and (row + row_adj) < main_filled.shape[0] and \
                        (col + col_adj) >= 0 and (col + col_adj) < main_filled.shape[1] and \
                        main_filled[row + row_adj, col + col_adj] != nodata_value and \
                        mask[row + row_adj, col + col_adj] == uint8_1:

                        weight = weights[idx]
                        value = main_filled[row + row_adj, col + col_adj]

                        value_sum += value * weight
                        weights_sum += weight
                        count += 1

                if count == 0:
                    local_filled[row, col] = nodata_value
                else:
                    local_filled[row, col] = value_sum * (1.0 / weights_sum)

        main_filled = local_filled
        iterations += 1

        if max_iterations is not None and iterations >= max_iterations:
            break

        if np.sum((main_filled == nodata_value) & (mask == uint8_1)) == 0:
            break

    return np.expand_dims(main_filled, axis=2)

@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True, inline='always')
def single_hue_to_rgb(
    p: float,
    q: float,
    t: float,
) -> float:
    """
    Helper function to convert hue to RGB.

    Args:
        p (float): Intermediate value used for hue to RGB conversion.
        q (float): Intermediate value used for hue to RGB conversion.
        t (float): Hue value.

    Returns:
        float: RGB value.
    """
    if t < 0:
        t += 1
    if t > 1:
        t -= 1
    if t < 1/6:
        return p + (q - p) * 6 * t
    if t < 1/2:
        return q
    if t < 2/3:
        return p + (q - p) * (2/3 - t) * 6
    return p


@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True, inline='always')
def single_hsl_to_rgb(
    h: float,
    s: float,
    l: float,
) -> Tuple[float, float, float]:
    """
    Convert a single HSL color to RGB.

    Args:
        h (float): Hue component.
        s (float): Saturation component.
        l (float): Lightness component.

    Returns:
        Tuple[float, float, float]: Tuple of RGB values (r, g, b).
    """
    if s == 0:
        return l, l, l

    q = l * (1 + s) if l < 0.5 else l + s - l * s
    p = 2 * l - q

    r = single_hue_to_rgb(p, q, h + 1/3)
    g = single_hue_to_rgb(p, q, h)
    b = single_hue_to_rgb(p, q, h - 1/3)

    return r, g, b


@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def hsl_to_rgb(hsl_array: np.ndarray) -> np.ndarray:
    """
    Convert an HSL array to an RGB array.

    Args:
        hsl_array (np.ndarray): Input HSL array with shape (height, width, 3).

    Returns:
        np.ndarray: Output RGB array with shape (height, width, 3).
    """
    assert hsl_array.ndim == 3, "Input array must have 3 dimensions"
    assert hsl_array.shape[-1] == 3, "Input array must have 3 channels"
    assert hsl_array.min() >= 0 and hsl_array.max() <= 1, "Input array must be normalized"

    shape = hsl_array.shape

    rgb_array = np.empty(shape, dtype=np.float32)
    for i in prange(shape[0]):
        for j in range(shape[1]):
            if hsl_array.ndim == 3:
                h, s, l = hsl_array[i, j, 0]
            else:
                h, s, l = hsl_array[i, j]

            r, g, b = single_hsl_to_rgb(h, s, l)

            if hsl_array.ndim == 3:
                rgb_array[i, j, 0] = r
                rgb_array[i, j, 1] = g
                rgb_array[i, j, 2] = b
            else:
                rgb_array[i, j] = [r, g, b]

    return rgb_array


def rgb_to_hsl(rgb_array: np.ndarray) -> np.ndarray:
    """
    Convert an RGB array to an HSL array.

    Args:
        rgb_array (np.ndarray): Input RGB array with shape (height, width, 3).

    Returns:
        np.ndarray: Output HSL array with shape (height, width, 3).
    """
    assert rgb_array.ndim == 3, "Input array must have 3 dimensions"
    assert rgb_array.shape[-1] == 3, "Input array must have 3 channels"
    assert rgb_array.min() >= 0 and rgb_array.max() <= 1, "Input array must be normalized"

    # Calculate the minimum, maximum, and difference of the RGB values
    cmin = np.min(rgb_array, axis=-1)
    cmax = np.max(rgb_array, axis=-1)
    delta = cmax - cmin

    # Calculate the HSL values
    h = np.zeros_like(delta)
    s = np.zeros_like(delta)
    l = (cmax + cmin) / 2

    mask = delta != 0
    s[mask] = delta[mask] / (1 - np.abs(2 * l[mask] - 1))

    r, g, b = rgb_array[..., 0], rgb_array[..., 1], rgb_array[..., 2]

    mask_r = (cmax == r) & mask
    mask_g = (cmax == g) & mask
    mask_b = (cmax == b) & mask

    h[mask_r] = np.divide(g[mask_r] - b[mask_r], delta[mask_r], out=np.zeros_like(delta[mask_r]), where=delta[mask_r] != 0) % 6
    h[mask_g] = np.divide(b[mask_g] - r[mask_g], delta[mask_g], out=np.zeros_like(delta[mask_g]), where=delta[mask_g] != 0) + 2
    h[mask_b] = np.divide(r[mask_b] - g[mask_b], delta[mask_b], out=np.zeros_like(delta[mask_b]), where=delta[mask_b] != 0) + 4

    h = (h * 60) % 360
    h[h < 0] += 360

    # Normalize the hue value to [0, 1]
    h /= 360

    hsl_array = np.stack([h, s, l], axis=-1)
    hsl_array = np.clip(hsl_array, 0.0, 1.0)

    return hsl_array



"""
### Reproject rasters. ###

Module to reproject rasters to a target coordinate reference system.
Can uses references from vector or other raster datasets.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
from osgeo import gdal, ogr, osr

# Internal
from buteo.utils import core_utils, gdal_utils, gdal_enums
from buteo.raster import core_raster


def find_common_projection(
    rasters: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
):
    """
        Find the common projection of a list of rasters. If no rasters have 
        the majority of the same projection, the function will return the
        projection of the first raster. If only on raster is provided, the
        projection of that raster will be returned.
    
        Args:
            rasters (list): A list of rasters.

        Returns:
            osr.SpatialReference: The common projection.
    """
    core_utils.type_check(rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "rasters")

    rasters = core_utils.ensure_list(rasters)

    assert gdal_utils.is_raster_list(rasters), "rasters must be a list of rasters."

    # Get the projection of each raster
    projections = [gdal_utils.parse_projection(raster) for raster in rasters]

    # Get the most common projection
    common_projection = max(set(projections), key=projections.count)

    return common_projection


def _reproject_raster(
    raster,
    projection,
    out_path=None,
    *,
    resample_alg="nearest",
    copy_if_same=True,
    overwrite=True,
    creation_options=None,
    dst_nodata="infer",
    dtype=None,
    prefix="",
    suffix="",
    add_uuid=False,
):
    """ Internal. """
    assert isinstance(raster, (gdal.Dataset, str)), f"The input raster must be in the form of a str or a gdal.Dataset: {raster}"

    out_path = gdal_utils.create_output_path(
        raster,
        out_path=out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
    )

    ref = core_raster._open_raster(raster)
    metadata = core_raster._raster_to_metadata(ref)

    out_format = gdal_utils.path_to_driver_raster(out_path)

    original_projection = gdal_utils.parse_projection(ref)
    target_projection = gdal_utils.parse_projection(projection)

    if not isinstance(original_projection, osr.SpatialReference):
        raise RuntimeError("Error while parsing input projection.")

    if not isinstance(target_projection, osr.SpatialReference):
        raise RuntimeError("Error while parsing target projection.")

    # The input is already in the correct projection.
    if not copy_if_same and gdal_utils.projections_match(original_projection, target_projection):
        return gdal_utils.get_path_from_dataset(ref)

    src_nodata = metadata["nodata_value"]
    out_nodata = None
    if dst_nodata == "infer":
        dst_nodata = src_nodata
    else:
        assert isinstance(dst_nodata, (int, float, type(None))), "dst_nodata must be an int, float, 'infer', or 'None'"
        out_nodata = dst_nodata

    if dtype is None:
        dtype = metadata["datatype"]

    core_utils.remove_if_required(out_path, overwrite)

    reprojected = gdal.Warp(
        out_path,
        ref,
        format=out_format,
        srcSRS=original_projection,
        dstSRS=target_projection,
        resampleAlg=gdal_enums.translate_resample_method(resample_alg),
        outputType=gdal_enums.translate_str_to_gdal_dtype(dtype),
        creationOptions=gdal_utils.default_creation_options(creation_options),
        srcNodata=src_nodata,
        dstNodata=out_nodata,
        multithread=True,
    )

    if reprojected is None:
        raise RuntimeError(f"Error while reprojecting raster: {raster}")

    return out_path


def reproject_raster(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    projection: Union[int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference],
    out_path: Optional[Union[str, List[str]]] = None,
    *,
    resample_alg: str = "nearest",
    copy_if_same: bool = True,
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
    dst_nodata: Union[str, int, float] = "infer",
    dtype: Optional[str] = None,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
) -> Union[str, List[str]]:
    """
    Reproject a raster(s) to a target coordinate reference system.

    Args:
        raster (str/list/gdal.Dataset): The raster(s) to reproject.
        projection (int/str/gdal.Dataset/ogr.DataSource/osr.SpatialReference): The projection is inferred from
            the input. The input can be: WKT proj, EPSG proj, Proj, osr proj, or read
            from a vector or raster datasource either from path or in-memory.

    Keyword Args:
        out_path (str/list/None, default=None): The output path. If not provided, the output path is inferred from the input.
        resample_alg (str, default="nearest"): The resampling algorithm.
        copy_if_same (bool, default=True): If the input and output projections are the same, copy the input raster to the output path.
        overwrite (bool, default=True): If the output path already exists, overwrite it.
        creation_options (list/None, default=None): A list of creation options for the output raster.
        dst_nodata (str/int/float, default="infer"): The nodata value for the output raster.
        dtype (str/None, default=None): The data type for the output raster.
        prefix (str, default=""): The prefix to add to the output path.
        suffix (str, default=""): The suffix to add to the output path.
        add_uuid (bool, default=False): If True, add a UUID to the output path.

    Returns:
        str/list: The output path(s).
    """
    core_utils.type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    core_utils.type_check(projection, [int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference], "projection")
    core_utils.type_check(out_path, [list, str, None], "out_path")
    core_utils.type_check(resample_alg, [str], "resample_alg")
    core_utils.type_check(overwrite, [bool], "overwrite")
    core_utils.type_check(creation_options, [[str], None], "creation_options")
    core_utils.type_check(dst_nodata, [str, int, float], "dst_nodata")
    core_utils.type_check(dtype, [str, None], "dtype")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "postfix")
    core_utils.type_check(add_uuid, [bool], "add_uuid")

    if core_utils.is_str_a_glob(raster):
        raster = core_utils.parse_glob_path(raster)

    raster_list = core_utils.ensure_list(raster)

    assert gdal_utils.is_raster_list(raster_list), f"The input raster(s) contains invalid elements: {raster_list}"

    path_list = gdal_utils.create_output_path_list(
        raster_list,
        out_path=out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
    )

    output = []
    for index, in_raster in enumerate(raster_list):
        output.append(
            _reproject_raster(
                in_raster,
                projection,
                out_path=path_list[index],
                resample_alg=resample_alg,
                copy_if_same=copy_if_same,
                overwrite=overwrite,
                creation_options=creation_options,
                dst_nodata=dst_nodata,
                dtype=dtype,
                prefix=prefix,
                suffix=suffix,
            )
        )

    if isinstance(raster, list):
        return output

    return output[0]


def match_raster_projections(
    rasters: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    reference: Union[str, gdal.Dataset, ogr.DataSource],
    *,
    out_path: Optional[Union[str, List[str]]] = None,
    overwrite: bool = True,
    dst_nodata: Union[str, int, float] = "infer",
    copy_if_already_correct: bool = True,
    creation_options: Optional[List[str]] = None,
) -> List[str]:
    """
    Match a raster or list of rasters to a master layer. The master can be
    either an OGR layer or a GDAL layer.

    Args:
        rasters (list): A list of rasters to match.
        reference (str/gdal.Dataset/ogr.DataSource): Path to the reference raster
            or vector.

    Keyword Args:
        out_path (str/list, default=None): Paths to the output. If not provided,
            the output will be in-memory rasters.
        overwrite (bool, default=True): If True, existing rasters will be
            overwritten.
        dst_nodata (str, default='infer'): Value to use for no-data pixels. If not
            provided, the value will be transfered from the original.
        copy_if_already_correct (bool, default=True): If True, the raster will be
            copied if it is already in the correct projection.
        creation_options (list, default=None): List of creation options to pass
            to the output raster.

    Returns:
        list: A list of reprojected input rasters with the correct projection.
    """
    core_utils.type_check(rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "rasters")
    core_utils.type_check(reference, [str, gdal.Dataset, ogr.DataSource], "reference")
    core_utils.type_check(out_path, [str, list, None], "out_path")
    core_utils.type_check(overwrite, [bool], "overwrite")
    core_utils.type_check(dst_nodata, [str, int, float], "dst_nodata")
    core_utils.type_check(copy_if_already_correct, [bool], "copy_if_already_correct")
    core_utils.type_check(creation_options, [[str], None], "creation_options")

    rasters = core_utils.ensure_list(rasters)

    assert gdal_utils.is_raster_list(rasters), "rasters must be a list of rasters."

    try:
        target_projection = gdal_utils.parse_projection(reference)
    except Exception:
        raise ValueError(f"Unable to parse projection from master. Received: {reference}") from None

    add_uuid = out_path is None

    path_list = gdal_utils.create_output_path_list(rasters, out_path, overwrite=overwrite, add_uuid=add_uuid, ext=".tif")

    output = []

    for index, in_raster in enumerate(rasters):
        path = _reproject_raster(
            in_raster,
            target_projection,
            out_path=path_list[index],
            overwrite=overwrite,
            copy_if_same=copy_if_already_correct,
            dst_nodata=dst_nodata,
            creation_options=gdal_utils.default_creation_options(creation_options),
        )

        output.append(path)

    return output



"""
### Merge vectors. ###

Merges vectors into a single vector file.
"""

# Standard library
import sys; sys.path.append("../../")

# External
from osgeo import ogr

# Internal
from buteo.utils import core_utils, gdal_utils
from buteo.vector import core_vector



def merge_vectors(
    vectors,
    out_path=None,
    *,
    preserve_fid=True,
):
    """
    Merge vectors to a single geopackage.

    ## Args:
    `vectors` (_list_): List of vectors to merge.

    ## Kwargs:
    `out_path` (_str_): Path to output vector. (Default: **None**) </br>
    `preserve_fid` (_bool_): Preserve FIDs? (Default: **True**)

    ## Returns:
    (_str_): Path to output vector.
    """
    core_utils.type_check(vectors, [[str, ogr.DataSource]], "vector")
    core_utils.type_check(out_path, [str, [str], None], "out_path")
    core_utils.type_check(preserve_fid, [bool], "preserve_fid")

    vector_list = core_utils.ensure_list(vectors)

    assert gdal_utils.is_vector_list(vector_list), "Invalid input vector list"

    if out_path is None:
        out_path = gdal_utils.create_memory_path(
            gdal_utils.get_path_from_dataset(vector_list[0]),
            prefix="",
            suffix="_merged",
            add_uuid=True,
        )

    driver = ogr.GetDriverByName(gdal_utils.path_to_driver_vector(out_path))

    merged_ds = driver.CreateDataSource(out_path)

    for vector in vector_list:
        ref = core_vector._open_vector(vector)
        metadata = core_vector._vector_to_metadata(ref)

        for layer in metadata["layers"]:
            name = layer["layer_name"]
            merged_ds.CopyLayer(ref.GetLayer(name), name, ["OVERWRITE=YES"])

    merged_ds.FlushCache()

    return out_path



"""
### GDAL Warp functions  ###

Module to wrap the functionality of GDAL's gdalwarp.
"""

# Standard library
import sys; sys.path.append("../../")
from uuid import uuid4

# External
from osgeo import gdal, ogr, osr

# Internal
from buteo.utils import core_utils, gdal_utils, bbox_utils, gdal_enums
from buteo.raster import core_raster
from buteo.vector import core_vector



def _warp_raster(
    raster,
    out_path=None,
    *,
    projection=None,
    clip_geom=None,
    target_size=None,
    target_in_pixels=False,
    resample_alg="nearest",
    crop_to_geom=True,
    all_touch=False,
    adjust_bbox=False,
    src_nodata="infer",
    dst_nodata="infer",
    layer_to_clip=0,
    prefix="",
    suffix="_resampled",
    overwrite=True,
    creation_options=None,
):
    """ Internal. """
    if out_path is None:
        out_path = gdal_utils.create_memory_path(
            out_path,
            prefix=prefix,
            suffix=suffix,
            add_uuid=True,
        )

    origin = core_raster._open_raster(raster)
    raster_metadata = core_raster._raster_to_metadata(origin)

    # options
    warp_options = []
    if all_touch:
        warp_options.append("CUTLINE_ALL_TOUCHED=TRUE")
    else:
        warp_options.append("CUTLINE_ALL_TOUCHED=FALSE")

    origin_projection = raster_metadata["projection_osr"]
    origin_extent = raster_metadata["get_bbox_as_vector_latlng"]() # pylint: disable=not-callable

    target_projection = origin_projection
    if projection is not None:
        target_projection = gdal_utils.parse_projection(projection)

    if clip_geom is not None:
        if gdal_utils.is_raster(clip_geom):
            opened_raster = core_raster._open_raster(clip_geom)
            clip_metadata_raster = core_raster._raster_to_metadata(opened_raster)
            clip_ds = clip_metadata_raster["get_bbox_as_vector"]() # pylint: disable=not-callable
            clip_metadata = core_vector._vector_to_metadata(clip_ds)
        elif gdal_utils.is_vector(clip_geom):
            clip_ds = core_vector.open_vector(clip_geom)
            clip_metadata = core_vector._vector_to_metadata(clip_ds)
        else:
            if core_utils.file_exists(clip_geom):
                raise ValueError(f"Unable to parse clip geometry: {clip_geom}")
            else:
                raise ValueError(f"Unable to find clip geometry {clip_geom}")

        if layer_to_clip > (clip_metadata["layer_count"] - 1):
            raise ValueError("Requested an unable layer_to_clip.")

        clip_projection = clip_metadata["projection_osr"]
        clip_extent = clip_metadata["get_bbox_as_vector_latlng"]() # pylint: disable=not-callable

        # Fast check: Does the extent of the two inputs overlap?
        if not origin_extent.Intersects(clip_extent):
            raise RuntimeError("Clipping geometry did not intersect raster.")

        # Check if projections match, otherwise reproject target geom.
        if not target_projection.IsSame(clip_projection):
            clip_metadata["bbox"] = bbox_utils.reproject_bbox(
                clip_metadata["bbox"],
                clip_projection,
                target_projection,
            )

        # The extent needs to be reprojected to the target.
        # this ensures that adjust_bbox works.
        output_bounds = bbox_utils.reproject_bbox(
            raster_metadata["extent"],
            origin_projection,
            target_projection,
        )

        if crop_to_geom:

            if adjust_bbox:
                output_bounds = bbox_utils.align_bboxes_to_pixel_size(
                    raster_metadata["bbox"],
                    clip_metadata["bbox"],
                    raster_metadata["pixel_width"],
                    raster_metadata["pixel_height"],
                )

            else:
                output_bounds = clip_metadata["bbox"]

        if clip_metadata["layer_count"] > 1:
            clip_ds = core_vector._vector_to_memory(
                clip_ds,
                memory_path=f"clip_geom_{uuid4().int}.gpkg",
                layer_to_extract=layer_to_clip,
            )
        elif not isinstance(clip_ds, str):
            clip_ds = core_vector._vector_to_memory(
                clip_ds,
                memory_path=f"clip_geom_{uuid4().int}.gpkg",
            )

        if clip_ds is None:
            raise ValueError(f"Unable to parse input clip geom: {clip_geom}")

    x_res, y_res, x_pixels, y_pixels = gdal_utils.parse_raster_size(
        target_size, target_in_pixels=target_in_pixels
    )

    out_format = gdal_utils.path_to_driver_raster(out_path)

    # nodata
    out_nodata = None
    if src_nodata is not None:
        out_nodata = raster_metadata["nodata_value"]
    else:
        if dst_nodata == "infer":
            out_nodata = raster_metadata["nodata_value"]
        else:
            out_nodata = dst_nodata

    # Removes file if it exists and overwrite is True.
    core_utils.remove_if_required(out_path, overwrite)

    warped = gdal.Warp(
        out_path,
        origin,
        xRes=x_res,
        yRes=y_res,
        width=x_pixels,
        height=y_pixels,
        cutlineDSName=clip_ds,
        outputBounds=bbox_utils.convert_geom_to_bbox(output_bounds),
        format=out_format,
        srcSRS=origin_projection,
        dstSRS=target_projection,
        resampleAlg=gdal_enums.translate_resample_method(resample_alg),
        creationOptions=gdal_utils.default_creation_options(creation_options),
        warpOptions=warp_options,
        srcNodata=src_nodata,
        dstNodata=out_nodata,
        targetAlignedPixels=False,
        cropToCutline=False,
        multithread=True,
    )

    if warped is None:
        raise RuntimeError(f"Error while warping raster: {raster}")

    return out_path


def warp_raster(
    raster,
    out_path=None,
    *,
    projection=None,
    clip_geom=None,
    target_size=None,
    target_in_pixels=False,
    resample_alg="nearest",
    crop_to_geom=True,
    all_touch=False,
    adjust_bbox=False,
    src_nodata="infer",
    dst_nodata="infer",
    layer_to_clip=0,
    prefix="",
    suffix="_resampled",
    add_uuid=False,
    overwrite=True,
    creation_options=None,
):
    """
    Warps a raster into a target raster. </br>

    Please be aware that all_touch does not work if target_size is set.
    If all_touch is required while resampling. Do it in two steps:
    resample -> warp or resample -> clip.

    ## Args:
    `raster` (_str_/_list_/_gdal.Dataset_): The raster(s) to warp.

    ## Kwargs:
    `out_path` (_str_/_list_/_None_): The path to the output raster(s). If not set, a memory raster is created. (Default: **None**) </br>
    `projection` (_str_/_osr.SpatialReference_/_gdal.Dataset_/_ogr.DataSource_): The projection of the output raster. If not set, the projection of the input raster is used. (Default: **None**) </br>
    `clip_geom` (_str_/_gdal.Dataset_/_ogr.DataSource_): The geometry to clip the raster to. (Default: **None**) </br>
    `target_size` (_tuple_/_None_): The target size of the output raster. If not set, the size of the input raster is used. (Default: **None**) </br>
    `target_in_pixels` (_bool_): If True, the target size is in pixels. If False, the target size is in map units. (Default: **False**) </br>
    `resample_alg` (_str_): The resampling algorithm. (Default: **nearest**) </br>
    `crop_to_geom` (_bool_): If True, the output raster is cropped to the extent of the clip geometry. (Default: **True**) </br>
    `all_touch` (_bool_): If True, all pixels touching the clipping geometry is included. (Default: **False**) </br>
    `adjust_bbox` (_bool_): If True, the bounding box of the output raster is adjusted to align with the clipping geometry. (Default: **False**) </br>
    `src_nodata` (_str_/_int_/_float_/_None_): The nodata value of the input raster. If not set, the nodata value of the input raster is used. (Default: **infer**) </br>
    `dst_nodata` (_str_/_int_/_float_/_None_): The nodata value of the output raster. If not set, the nodata value of the input raster is used. (Default: **infer**) </br>
    `layer_to_clip` (_int_): The layer to clip the raster to. (Default: **0**) </br>
    `prefix` (_str_): The prefix of the output raster. (Default: **""**) </br>
    `suffix` (_str_): The suffix of the output raster. (Default: **"_resampled"**) </br>
    `add_uuid` (_bool_): If True, a unique identifier is added to the output raster. (Default: **False**) </br>
    `overwrite` (_bool_): If True, the output raster is overwritten if it already exists. (Default: **True**) </br>
    `creation_options` (_list_): The creation options of the output raster. (Default: **None**) </br>

    ## Returns:
    (_str_/_list_): The path to the warped output raster(s).
    """
    core_utils.type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    core_utils.type_check(out_path, [str, [str], None], "out_path")
    core_utils.type_check(projection, [int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference, None], "projection")
    core_utils.type_check(clip_geom, [str, ogr.DataSource], "clip_geom")
    core_utils.type_check(target_size, [tuple, list, int, float, str, gdal.Dataset, None], "target_size")
    core_utils.type_check(target_in_pixels, [bool], "target_in_pixels")
    core_utils.type_check(resample_alg, [str], "resample_alg")
    core_utils.type_check(crop_to_geom, [bool], "crop_to_geom")
    core_utils.type_check(all_touch, [bool], "all_touch")
    core_utils.type_check(adjust_bbox, [bool], "adjust_bbox")
    core_utils.type_check(src_nodata, [str, int, float], "src_nodata")
    core_utils.type_check(dst_nodata, [str, int, float], "dst_nodata")
    core_utils.type_check(layer_to_clip, [int], "layer_to_clip")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "postfix")
    core_utils.type_check(add_uuid, [bool], "add_uuid")
    core_utils.type_check(overwrite, [bool], "overwrite")
    core_utils.type_check(creation_options, [[str], None], "creation_options")

    raster_list = core_utils.ensure_list(raster)
    path_list = gdal_utils.create_output_path_list(
        raster_list,
        out_path=out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        overwrite=overwrite,
    )

    output = []
    for index, in_raster in enumerate(raster_list):
        output.append(
            _warp_raster(
                in_raster,
                out_path=path_list[index],
                projection=projection,
                clip_geom=clip_geom,
                target_size=target_size,
                target_in_pixels=target_in_pixels,
                resample_alg=resample_alg,
                crop_to_geom=crop_to_geom,
                all_touch=all_touch,
                adjust_bbox=adjust_bbox,
                overwrite=overwrite,
                creation_options=creation_options,
                src_nodata=src_nodata,
                dst_nodata=dst_nodata,
                layer_to_clip=layer_to_clip,
                prefix=prefix,
                suffix=suffix,
            )
        )

    if isinstance(raster, list):
        return output

    return output[0]



"""
### Handle and create borders on rasters. ###

Functions to add or remove borders from rasters.
Useful for warped satellite images and for proximity searching.
"""

# TODO: Remove near black borders.

# Standard library
import sys; sys.path.append("../../")

# External
import numpy as np
from osgeo import gdal

# Internal
from buteo.raster import core_raster
from buteo.utils import gdal_utils, core_utils, gdal_enums


def _add_border_to_raster(
    raster,
    out_path=None,
    border_size=100,
    *,
    border_size_unit="px",
    border_value=0,
    overwrite=True,
    creation_options=None,
):
    """ Internal. """
    in_raster = core_raster.open_raster(raster)
    metadata = core_raster.raster_to_metadata(in_raster)

    # Parse the driver
    driver_name = "GTiff" if out_path is None else gdal_utils.path_to_driver_raster(out_path)
    if driver_name is None:
        raise ValueError(f"Unable to parse filetype from path: {out_path}")

    driver = gdal.GetDriverByName(driver_name)
    if driver is None:
        raise ValueError(f"Error while creating driver from extension: {out_path}")

    output_name = None
    if out_path is None:
        output_name = gdal_utils.create_memory_path("raster_proximity.tif", add_uuid=True)
    else:
        output_name = out_path

    in_arr = core_raster.raster_to_array(in_raster)

    if border_size_unit == "px":
        border_size_y = border_size
        border_size_x = border_size
        new_shape = (
            in_arr.shape[0] + (2 * border_size_y),
            in_arr.shape[1] + (2 * border_size_x),
            in_arr.shape[2],
        )
    else:
        border_size_y = round(border_size / metadata["pixel_height"])
        border_size_x = round(border_size / metadata["pixel_width"])
        new_shape = (
            in_arr.shape[0] + (2 * border_size_y),
            in_arr.shape[1] + (2 * border_size_x),
            in_arr.shape[2],
        )

    new_arr = np.full(new_shape, border_value, dtype=in_arr.dtype)
    new_arr[border_size_y:-border_size_y, border_size_x:-border_size_x, :] = in_arr

    if isinstance(in_arr, np.ma.MaskedArray):
        mask = np.zeros(new_shape, dtype=bool)
        mask[
            border_size_y:-border_size_y, border_size_x:-border_size_x, :
        ] = in_arr.mask
        new_arr = np.ma.array(new_arr, mask=mask)
        new_arr.fill_value = in_arr.fill_value

    core_utils.remove_if_required(out_path, overwrite)

    dest_raster = driver.Create(
        output_name,
        new_shape[1],
        new_shape[0],
        metadata["band_count"],
        gdal_enums.translate_str_to_gdal_dtype(in_arr.dtype),
        gdal_utils.default_creation_options(creation_options),
    )

    og_transform = in_raster.GetGeoTransform()

    new_transform = []
    for i in og_transform:
        new_transform.append(i)

    new_transform[0] -= border_size_x * og_transform[1]
    new_transform[3] -= border_size_y * og_transform[5]

    dest_raster.SetGeoTransform(new_transform)
    dest_raster.SetProjection(in_raster.GetProjectionRef())

    for band_num in range(1, metadata["band_count"] + 1):
        dst_band = dest_raster.GetRasterBand(band_num)
        dst_band.WriteArray(new_arr[:, :, band_num - 1])

        if metadata["has_nodata"]:
            dst_band.SetNoDataValue(metadata["nodata_value"])

    return output_name


def add_border_to_raster(
    raster,
    out_path=None,
    border_size=100,
    *,
    border_size_unit="px",
    border_value=0,
    allow_lists=True,
    overwrite=True,
    creation_options=None,
):
    """
    Add a border to a raster.

    Args:
        input_raster (str/gdal.DataSet): The input raster.

    Keyword Args:
        out_path (str/None, default=None): The output path. If None, the output
            will be a memory raster.
        border_size (int, default=100): The size of the border.
        border_size_unit (str, default='px'): The unit of the border size.
        border_value (int, default=0): The value of the border.
        overwrite (bool, default=True): If True, the output raster will be
            overwritten.
        allow_lists (bool, default=True): If True, lists of rasters will be
            allowed.
        creation_options (list/None, default=None): Creation options for the
            output raster.

    Returns:
        str/gdal.DataSet: The output raster with added borders.
    """
    core_utils.type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    core_utils.type_check(out_path, [str, None], "out_path")
    core_utils.type_check(border_size, [int], "border_size")
    core_utils.type_check(border_size_unit, [str], "border_size_unit")
    core_utils.type_check(border_value, [int, float], "border_value")
    core_utils.type_check(overwrite, [bool], "overwrite")
    core_utils.type_check(creation_options, [list, None], "creation_options")

    if not allow_lists:
        if isinstance(raster, list):
            raise ValueError("Lists are not allowed as input.")

        return _add_border_to_raster(
            raster,
            out_path=out_path,
            border_size=border_size,
            border_size_unit=border_size_unit,
            border_value=border_value,
            overwrite=overwrite,
            creation_options=creation_options,
        )

    raster_list = core_utils.ensure_list(raster)

    if out_path is None:
        out_path = gdal_utils.create_memory_path("raster_proximity.tif", add_uuid=True)

    for raster_path in raster_list:
        _add_border_to_raster(
            raster_path,
            out_path=out_path,
            border_size=border_size,
            border_size_unit=border_size_unit,
            border_value=border_value,
            overwrite=overwrite,
            creation_options=creation_options,
        )

    return out_path



"""
### Perform morphological operations on arrays and rasters.  ###
"""

# Standard library
import sys; sys.path.append("../../")

# External
import numpy as np

# Internal
from buteo.raster.convolution import convolve_array, get_kernel, pad_array_view


DEFAULT_SPHERICAL = False
DEFAULT_WEIGHT = None
DEFAULT_DECAY = 0.2
DEFAULT_SIGMA = 1.0


def adjust_kernel(
    kernel,
    spherical=DEFAULT_SPHERICAL,
    distance_weight=DEFAULT_WEIGHT,
    distance_decay=DEFAULT_DECAY,
    distance_sigma=DEFAULT_SIGMA,
):
    """ Adjust an edge kernel by a normal kernel. Only works for symmetric kernels."""
    normal_kernel, _weights, _offsets = get_kernel(
        kernel.shape[0],
        depth=1,
        normalise=False,
        spherical=spherical,
        distance_weight=distance_weight,
        distance_decay=distance_decay,
        distance_sigma=distance_sigma,
    )

    if spherical is True or distance_weight is not None:
        kernel_pre = np.abs(kernel).sum()

        kernel = kernel * normal_kernel

        kernel_post = np.abs(kernel).sum()

        kernel = kernel * (kernel_pre / kernel_post)

    return kernel


def get_sobel_kernel(
    size=3,
    scale=1,
    spherical=DEFAULT_SPHERICAL,
    distance_weight=DEFAULT_WEIGHT,
    distance_decay=DEFAULT_DECAY,
    distance_sigma=DEFAULT_SIGMA,
):
    """ Get a sobel kernel of arbitrary size. """

    assert size % 2 != 0

    gx = np.zeros((size, size, 1), dtype=np.float32)
    gy = np.zeros((size, size, 1), dtype=np.float32)

    indices = np.indices((size, size, 1), dtype=np.float32)
    cols = indices[0] - (size // 2)
    rows = indices[1] - (size // 2)

    squared = np.power(cols, 2) + np.power(rows, 2)
    np.divide(cols, squared, out=gy, where=squared!=0) # in-place
    np.divide(rows, squared, out=gx, where=squared!=0) # in-place

    gx = gx * scale
    gy = gy * scale

    gx = adjust_kernel(
        gx,
        spherical=spherical,
        distance_weight=distance_weight,
        distance_decay=distance_decay,
        distance_sigma=distance_sigma,
    )
    gy = adjust_kernel(
        gy,
        spherical=spherical,
        distance_weight=distance_weight,
        distance_decay=distance_decay,
        distance_sigma=distance_sigma,
    )

    weights_x = []
    weights_y = []

    offsets_x = []
    offsets_y = []

    for idx_x in range(0, gx.shape[0]):
        for idx_y in range(0, gx.shape[1]):
            for idx_z in range(0, gx.shape[2]):
                current_weight_x = gx[idx_x][idx_y][idx_z]
                current_weight_y = gy[idx_x][idx_y][idx_z]

                if current_weight_x != 0.0:
                    offsets_x.append(
                        [
                            idx_x - (gx.shape[0] // 2),
                            idx_y - (gx.shape[1] // 2),
                            idx_z,
                        ]
                    )

                    weights_x.append(current_weight_x)

                if current_weight_y != 0.0:
                    offsets_y.append(
                        [
                            idx_x - (gy.shape[0] // 2),
                            idx_y - (gy.shape[1] // 2),
                            idx_z,
                        ]
                    )

                    weights_y.append(current_weight_y)

    return (
        (gx.astype("float32"), np.array(weights_x, dtype="float32"), np.array(offsets_x, dtype=int)),
        (gy.astype("float32"), np.array(weights_y, dtype="float32"), np.array(offsets_y, dtype=int)),
    )


def get_roberts_prewitt_kernel(
    roberts=True,
    spherical=DEFAULT_SPHERICAL,
    distance_weight=DEFAULT_WEIGHT,
    distance_decay=DEFAULT_DECAY,
    distance_sigma=DEFAULT_SIGMA,
):
    """ Get a roberts cross or prewitt kernel, only available for size 2x2 and 3x3 """

    if roberts:
        gx = np.array([
            [ 0, 0, 0 ],
            [ 0, 1, 0 ],
            [ 0, 0,-1 ],
        ])[:, :, np.newaxis]

        gy = np.array([
            [ 0, 0, 0 ],
            [ 0, 0, 1 ],
            [ 0,-1, 0 ],
        ])[:, :, np.newaxis]
    else:
        gx = np.array([
            [-1, 0, 1 ],
            [-1, 0, 1 ],
            [-1, 0, 1 ],
        ])[:, :, np.newaxis]

        gy = np.array([
            [ 1, 1, 1 ],
            [ 0, 0, 0 ],
            [-1,-1,-1 ],
        ])[:, :, np.newaxis]

    gx = adjust_kernel(
        gx,
        spherical=spherical,
        distance_weight=distance_weight,
        distance_decay=distance_decay,
        distance_sigma=distance_sigma,
    )
    gy = adjust_kernel(
        gy,
        spherical=spherical,
        distance_weight=distance_weight,
        distance_decay=distance_decay,
        distance_sigma=distance_sigma,
    )

    weights_x = []
    weights_y = []

    offsets_x = []
    offsets_y = []

    for idx_x in range(0, gx.shape[0]):
        for idx_y in range(0, gx.shape[1]):
            for idx_z in range(0, gx.shape[2]):
                current_weight_x = gx[idx_x][idx_y][idx_z]
                current_weight_y = gy[idx_x][idx_y][idx_z]

                if current_weight_x != 0.0:
                    offsets_x.append(
                        [
                            idx_x - (gx.shape[0] // 2),
                            idx_y - (gx.shape[1] // 2),
                            idx_z,
                        ]
                    )

                    weights_x.append(current_weight_x)

                if current_weight_y != 0.0:
                    offsets_y.append(
                        [
                            idx_x - (gy.shape[0] // 2),
                            idx_y - (gy.shape[1] // 2),
                            idx_z,
                        ]
                    )

                    weights_y.append(current_weight_y)

    return (
        (gx.astype("float32"), np.array(weights_x, dtype="float32"), np.array(offsets_x, dtype=int)),
        (gy.astype("float32"), np.array(weights_y, dtype="float32"), np.array(offsets_y, dtype=int)),
    )


def edge_detection(
    arr,
    method,
    filter_size=3,
    spherical=DEFAULT_SPHERICAL,
    distance_weight=DEFAULT_WEIGHT,
    distance_decay=DEFAULT_DECAY,
    distance_sigma=DEFAULT_SIGMA,
    scale=1.0,
    merge_results=True,
    gradient_output=False,
    nodata=False,
    nodata_value=-9999.9,
):
    """ Perform an detection method. """

    if method == "sobel":
        gx, gy = get_sobel_kernel(
            size=filter_size,
            scale=scale,
            spherical=spherical,
            distance_weight=distance_weight,
            distance_decay=distance_decay,
            distance_sigma=distance_sigma,
        )
    elif method == "roberts":
        gx, gy = get_roberts_prewitt_kernel(
            roberts=True,
            spherical=spherical,
            distance_weight=distance_weight,
            distance_decay=distance_decay,
            distance_sigma=distance_sigma,
        )
    elif method == "prewitt":
        gx, gy = get_roberts_prewitt_kernel(
            roberts=False,
            spherical=spherical,
            distance_weight=distance_weight,
            distance_decay=distance_decay,
            distance_sigma=distance_sigma,
        )
    else:
        raise ValueError(f"Unknown method: {method}")

    _kernel_x, weights_x, offsets_x = gx
    _kernel_y, weights_y, offsets_y = gy

    pad_size = filter_size // 2
    arr_padded = pad_array_view(arr, pad_size)
    gx = convolve_array(arr_padded, offsets=offsets_x, weights=weights_x, method="sum", nodata=nodata, nodata_value=nodata_value)
    gx = gx[pad_size:-pad_size, pad_size:-pad_size, :]
    gy = convolve_array(arr_padded, offsets=offsets_y, weights=weights_y, method="sum", nodata=nodata, nodata_value=nodata_value)
    gy = gy[pad_size:-pad_size, pad_size:-pad_size, :]
    if gradient_output:
        grad_mag = np.sqrt(np.add(np.power(gx, 2), np.power(gy, 2)))
        grad_dir = np.arctan(np.true_divide(gx, gy, where=gy!=0))

        return grad_mag, grad_dir

    if merge_results:
        return np.sqrt(np.add(np.power(gx, 2), np.power(gy, 2)))

    return gx, gy


def edge_detection_sobel(arr, filter_size=3, spherical=DEFAULT_SPHERICAL, distance_weight=DEFAULT_WEIGHT, distance_decay=DEFAULT_DECAY, distance_sigma=DEFAULT_SIGMA, scale=1.0, merge_results=True, gradient_output=False, nodata=False, nodata_value=-9999.9):
    """ Detect edges using the sobel filtering method. """

    return edge_detection(
        arr,
        method="sobel",
        filter_size=filter_size,
        spherical=spherical,
        distance_weight=distance_weight,
        distance_decay=distance_decay,
        distance_sigma=distance_sigma,
        scale=scale,
        merge_results=merge_results,
        gradient_output=gradient_output,
        nodata=nodata,
        nodata_value=nodata_value,
    )


def edge_detection_roberts(arr, spherical=DEFAULT_SPHERICAL, distance_weight=DEFAULT_WEIGHT, distance_decay=DEFAULT_DECAY, distance_sigma=DEFAULT_SIGMA, merge_results=True, gradient_output=False, nodata=False, nodata_value=-9999.9):
    """ Detect edges using the roberts filtering method. """

    return edge_detection(
        arr,
        method="roberts",
        spherical=spherical,
        distance_weight=distance_weight,
        distance_decay=distance_decay,
        distance_sigma=distance_sigma,
        merge_results=merge_results,
        gradient_output=gradient_output,
        nodata=nodata,
        nodata_value=nodata_value,
    )


def edge_detection_prewitt(arr, spherical=DEFAULT_SPHERICAL, distance_weight=DEFAULT_WEIGHT, distance_decay=DEFAULT_DECAY, distance_sigma=DEFAULT_SIGMA, merge_results=True, gradient_output=False, nodata=False, nodata_value=-9999.9):
    """ Detect edges using the prewitt filtering method. """

    return edge_detection(
        arr,
        method="prewitt",
        spherical=spherical,
        distance_weight=distance_weight,
        distance_decay=distance_decay,
        distance_sigma=distance_sigma,
        merge_results=merge_results,
        gradient_output=gradient_output,
        nodata=nodata,
        nodata_value=nodata_value,
    )
